# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for email settings
"""

import colander

from wuttaweb.views import MasterView
from wuttaweb.forms.schema import EmailRecipients


class EmailSettingView(MasterView):
    """
    Master view for :term:`email settings <email setting>`.
    """
    model_name = 'email_setting'
    model_key = 'key'
    model_title = "Email Setting"
    url_prefix = '/email/settings'
    filterable = False
    sortable = True
    sort_on_backend = False
    paginated = False
    creatable = False
    deletable = False

    labels = {
        'key': "Email Key",
        'replyto': "Reply-To",
    }

    grid_columns = [
        'key',
        'subject',
        'to',
        'enabled',
    ]

    # TODO: why does this not work?
    sort_defaults = 'key'

    form_fields = [
        'key',
        'description',
        'subject',
        'sender',
        'replyto',
        'to',
        'cc',
        'bcc',
        'notes',
        'enabled',
    ]

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.email_handler = self.app.get_email_handler()

    def get_grid_data(self, columns=None, session=None):
        """
        This view calls
        :meth:`~wuttjamaican:wuttjamaican.email.EmailHandler.get_email_settings()`
        on the :attr:`email_handler` to obtain its grid data.
        """
        data = []
        for setting in self.email_handler.get_email_settings().values():
            data.append(self.normalize_setting(setting))
        return data

    def normalize_setting(self, setting):
        """ """
        key = setting.__name__
        return {
            'key': key,
            'description': setting.__doc__,
            'subject': self.email_handler.get_auto_subject(key, rendered=False, setting=setting),
            'sender': self.email_handler.get_auto_sender(key),
            'replyto': self.email_handler.get_auto_replyto(key) or colander.null,
            'to': self.email_handler.get_auto_to(key),
            'cc': self.email_handler.get_auto_cc(key),
            'bcc': self.email_handler.get_auto_bcc(key),
            'notes': self.email_handler.get_notes(key) or colander.null,
            'enabled': self.email_handler.is_enabled(key),
        }

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)

        # key
        g.set_searchable('key')
        g.set_link('key')

        # subject
        g.set_searchable('subject')
        g.set_link('subject')

        # to
        g.set_renderer('to', self.render_to_short)

    def render_to_short(self, setting, field, value):
        """ """
        recips = value
        if not recips:
            return

        if len(recips) < 3:
            return ', '.join(recips)

        recips = ', '.join(recips[:2])
        return f"{recips}, ..."

    def get_instance(self):
        """ """
        key = self.request.matchdict['key']
        setting = self.email_handler.get_email_setting(key, instance=False)
        if setting:
            return self.normalize_setting(setting)

        raise self.notfound()

    def get_instance_title(self, setting):
        """ """
        return setting['subject']

    def configure_form(self, f):
        """ """
        super().configure_form(f)

        # description
        f.set_readonly('description')

        # replyto
        f.set_required('replyto', False)

        # to
        f.set_node('to', EmailRecipients())

        # cc
        f.set_node('cc', EmailRecipients())

        # bcc
        f.set_node('bcc', EmailRecipients())

        # notes
        f.set_widget('notes', 'notes')
        f.set_required('notes', False)

        # enabled
        f.set_node('enabled', colander.Boolean())

    def persist(self, setting):
        """ """
        session = self.Session()
        key = self.request.matchdict['key']

        def save(name, value):
            self.app.save_setting(session, f'{self.config.appname}.email.{key}.{name}', value)

        def delete(name):
            self.app.delete_setting(session, f'{self.config.appname}.email.{key}.{name}')

        # subject
        if setting['subject']:
            save('subject', setting['subject'])
        else:
            delete('subject')

        # sender
        if setting['sender']:
            save('sender', setting['sender'])
        else:
            delete('sender')

        # replyto
        if setting['replyto']:
            save('replyto', setting['replyto'])
        else:
            delete('replyto')

        # to
        if setting['to']:
            save('to', setting['to'])
        else:
            delete('to')

        # cc
        if setting['cc']:
            save('cc', setting['cc'])
        else:
            delete('cc')

        # bcc
        if setting['bcc']:
            save('bcc', setting['bcc'])
        else:
            delete('bcc')

        # notes
        if setting['notes']:
            save('notes', setting['notes'])
        else:
            delete('notes')

        # enabled
        save('enabled', 'true' if setting['enabled'] else 'false')

    def render_to_response(self, template, context):
        """ """
        if self.viewing:
            setting = context['instance']
            context['setting'] = setting
            context['has_html_template'] = self.email_handler.get_auto_body_template(
                setting['key'], 'html')
            context['has_txt_template'] = self.email_handler.get_auto_body_template(
                setting['key'], 'txt')

        return super().render_to_response(template, context)

    def preview(self):
        """
        View for showing a rendered preview of a given email template.

        This will render the email template according to the "mode"
        requested - i.e. HTML or TXT.
        """
        key = self.request.matchdict['key']
        setting = self.email_handler.get_email_setting(key)
        context = setting.sample_data()
        mode = self.request.params.get('mode', 'html')

        if mode == 'txt':
            body = self.email_handler.get_auto_txt_body(key, context)
            self.request.response.content_type = 'text/plain'

        else: # html
            body = self.email_handler.get_auto_html_body(key, context)

        self.request.response.text = body
        return self.request.response

    @classmethod
    def defaults(cls, config):
        """ """
        cls._email_defaults(config)
        cls._defaults(config)

    @classmethod
    def _email_defaults(cls, config):
        """ """
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_title_plural = cls.get_model_title_plural()
        instance_url_prefix = cls.get_instance_url_prefix()

        # fix permission group
        config.add_wutta_permission_group(permission_prefix,
                                          model_title_plural,
                                          overwrite=False)

        # preview
        config.add_route(f'{route_prefix}.preview',
                         f'{instance_url_prefix}/preview')
        config.add_view(cls, attr='preview',
                        route_name=f'{route_prefix}.preview',
                        permission=f'{permission_prefix}.view')


def defaults(config, **kwargs):
    base = globals()

    EmailSettingView = kwargs.get('EmailSettingView', base['EmailSettingView'])
    EmailSettingView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-

from unittest import TestCase

from pyramid import testing


class TestIncludeMe(TestCase):

    def test_basic(self):
        with testing.testConfig() as pyramid_config:

            # just ensure no error happens when included..
            pyramid_config.include('wuttaweb.static')



# -*- coding: utf-8; -*-

from unittest.mock import patch

from wuttaweb import handler as mod, static
from wuttaweb.forms import Form
from wuttaweb.grids import Grid
from wuttaweb.menus import MenuHandler
from wuttaweb.testing import WebTestCase


class MockMenuHandler(MenuHandler):
    pass

class LegacyMenuHandler(MenuHandler):
    pass

class AnotherMenuHandler(MenuHandler):
    pass


class TestWebHandler(WebTestCase):

    def make_handler(self):
        return mod.WebHandler(self.config)

    def test_get_fanstatic_url(self):
        handler = self.make_handler()

        # default with / root path
        url = handler.get_fanstatic_url(self.request, static.logo)
        self.assertEqual(url, '/fanstatic/wuttaweb_img/logo.png')

        # what about a subpath
        self.request.script_name = '/testing'
        url = handler.get_fanstatic_url(self.request, static.logo)
        self.assertEqual(url, '/testing/fanstatic/wuttaweb_img/logo.png')

    def test_get_favicon_url(self):
        handler = self.make_handler()

        # default
        url = handler.get_favicon_url(self.request)
        self.assertEqual(url, '/fanstatic/wuttaweb_img/favicon.ico')

        # config override
        self.config.setdefault('wuttaweb.favicon_url', '/testing/other.ico')
        url = handler.get_favicon_url(self.request)
        self.assertEqual(url, '/testing/other.ico')

    def test_get_header_logo_url(self):
        handler = self.make_handler()

        # default
        url = handler.get_header_logo_url(self.request)
        self.assertEqual(url, '/fanstatic/wuttaweb_img/favicon.ico')

        # config override
        self.config.setdefault('wuttaweb.header_logo_url', '/testing/header.png')
        url = handler.get_header_logo_url(self.request)
        self.assertEqual(url, '/testing/header.png')

    def test_get_main_logo_url(self):
        handler = self.make_handler()

        # default
        url = handler.get_main_logo_url(self.request)
        self.assertEqual(url, '/fanstatic/wuttaweb_img/logo.png')

        # config override
        self.config.setdefault('wuttaweb.logo_url', '/testing/other.png')
        url = handler.get_main_logo_url(self.request)
        self.assertEqual(url, '/testing/other.png')

    def test_get_menu_handler(self):
        handler = self.make_handler()

        # built-in default
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, MenuHandler)
        self.assertIs(type(menus), MenuHandler)

        # configured default
        self.config.setdefault('wutta.web.menus.handler.default_spec',
                               'tests.test_handler:MockMenuHandler')
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, MockMenuHandler)

        # configured handler (legacy)
        self.config.setdefault('wutta.web.menus.handler_spec',
                               'tests.test_handler:LegacyMenuHandler')
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, LegacyMenuHandler)

        # configued handler (proper)
        self.config.setdefault('wutta.web.menus.handler.spec',
                               'tests.test_handler:AnotherMenuHandler')
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, AnotherMenuHandler)

    def test_get_menu_handler_specs(self):
        handler = self.make_handler()

        # at least one spec by default
        specs = handler.get_menu_handler_specs()
        self.assertIn('wuttaweb.menus:MenuHandler', specs)

        # caller can specify default as string
        specs = handler.get_menu_handler_specs(default='tests.test_handler:MockMenuHandler')
        self.assertIn('wuttaweb.menus:MenuHandler', specs)
        self.assertIn('tests.test_handler:MockMenuHandler', specs)
        self.assertNotIn('tests.test_handler:AnotherMenuHandler', specs)

        # caller can specify default as list
        specs = handler.get_menu_handler_specs(default=[
            'tests.test_handler:MockMenuHandler',
            'tests.test_handler:AnotherMenuHandler'])
        self.assertIn('wuttaweb.menus:MenuHandler', specs)
        self.assertIn('tests.test_handler:MockMenuHandler', specs)
        self.assertIn('tests.test_handler:AnotherMenuHandler', specs)

        # default can be configured
        self.config.setdefault('wutta.web.menus.handler.default_spec',
                               'tests.test_handler:AnotherMenuHandler')
        specs = handler.get_menu_handler_specs()
        self.assertIn('wuttaweb.menus:MenuHandler', specs)
        self.assertNotIn('tests.test_handler:MockMenuHandler', specs)
        self.assertIn('tests.test_handler:AnotherMenuHandler', specs)

        # the rest come from entry points
        with patch.object(mod, 'load_entry_points', return_value={
                'legacy': LegacyMenuHandler,
        }):
            specs = handler.get_menu_handler_specs()
            self.assertNotIn('wuttaweb.menus:MenuHandler', specs)
            self.assertNotIn('tests.test_handler:MockMenuHandler', specs)
            self.assertIn('tests.test_handler:LegacyMenuHandler', specs)
            # nb. this remains from previous config default
            self.assertIn('tests.test_handler:AnotherMenuHandler', specs)

    def test_make_form(self):
        handler = self.make_handler()
        form = handler.make_form(self.request)
        self.assertIsInstance(form, Form)

    def test_make_grid(self):
        handler = self.make_handler()
        grid = handler.make_grid(self.request)
        self.assertIsInstance(grid, Grid)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Main Menu
"""

import re
import logging

from wuttjamaican.app import GenericHandler


log = logging.getLogger(__name__)


class MenuHandler(GenericHandler):
    """
    Base class and default implementation for :term:`menu handler`.

    It is assumed that most apps will override the menu handler with
    their own subclass.  In particular the subclass will override
    :meth:`make_menus()` and/or :meth:`make_admin_menu()`.

    The app should normally not instantiate the menu handler directly,
    but instead call
    :meth:`~wuttaweb.app.WebAppProvider.get_web_menu_handler()` on the
    :term:`app handler`.

    To configure your menu handler to be used, do this within your
    :term:`config extension`::

       config.setdefault('wuttaweb.menus.handler_spec', 'poser.web.menus:PoserMenuHandler')

    The core web app will call :meth:`do_make_menus()` to get the
    final (possibly filtered) menu set for the current user.  The
    menu set should be a list of dicts, for example::

       menus = [
           {
               'title': "First Dropdown",
               'type': 'menu',
               'items': [
                   {
                       'title': "Foo",
                       'route': 'foo',
                   },
                   {'type': 'sep'},     # horizontal line
                   {
                       'title': "Bar",
                       'route': 'bar',
                   },
               ],
           },
           {
               'title': "Second Dropdown",
               'type': 'menu',
               'items': [
                   {
                       'title': "Wikipedia",
                       'url': 'https://en.wikipedia.org',
                       'target': '_blank',
                   },
               ],
           },
       ]
    """

    ##############################
    # default menu definitions
    ##############################

    def make_menus(self, request, **kwargs):
        """
        Generate the full set of menus for the app.

        This method provides a semi-sane menu set by default, but it
        is expected for most apps to override it.

        The return value should be a list of dicts as described above.

        The default logic returns a list of menus obtained from
        calling these methods:

        * :meth:`make_people_menu()`
        * :meth:`make_admin_menu()`
        """
        return [
            self.make_people_menu(request),
            self.make_admin_menu(request),
        ]

    def make_people_menu(self, request, **kwargs):
        """
        Generate a typical People menu.

        This method provides a semi-sane menu set by default, but it
        is expected for most apps to override it.

        The return value for this method should be a *single* dict,
        which will ultimately be one element of the final list of
        dicts as described in :class:`MenuHandler`.
        """
        return {
            'title': "People",
            'type': 'menu',
            'items': [
                {
                    'title': "All People",
                    'route': 'people',
                    'perm': 'people.list',
                },
            ],
        }

    def make_admin_menu(self, request, **kwargs):
        """
        Generate a typical Admin menu.

        This method provides a semi-sane menu set by default, but it
        is expected for most apps to override it.

        The return value for this method should be a *single* dict,
        which will ultimately be one element of the final list of
        dicts as described in :class:`MenuHandler`.

        :param title: Override the menu title; default is "Admin".

        :param include_people: You can pass this flag to indicate the
           admin menu should contain an entry for the "People" view.
        """
        items = []

        if kwargs.get('include_people'):
            items.extend([
                {
                    'title': "All People",
                    'route': 'people',
                    'perm': 'people.list',
                },
            ])

        items.extend([
            {
                'title': "Users",
                'route': 'users',
                'perm': 'users.list',
            },
            {
                'title': "Roles",
                'route': 'roles',
                'perm': 'roles.list',
            },
            {
                'title': "Permissions",
                'route': 'permissions',
                'perm': 'permissions.list',
            },
            {'type': 'sep'},
            {
                'title': "Email Settings",
                'route': 'email_settings',
                'perm': 'email_settings.list',
            },
            {'type': 'sep'},
            {
                'title': "App Info",
                'route': 'appinfo',
                'perm': 'appinfo.list',
            },
            {
                'title': "Raw Settings",
                'route': 'settings',
                'perm': 'settings.list',
            },
            {
                'title': "Upgrades",
                'route': 'upgrades',
                'perm': 'upgrades.list',
            },
        ])

        return {
            'title': kwargs.get('title', "Admin"),
            'type': 'menu',
            'items': items,
        }

    ##############################
    # default internal logic
    ##############################

    def do_make_menus(self, request, **kwargs):
        """
        This method is responsible for constructing the final menu
        set.  It first calls :meth:`make_menus()` to get the basic
        set, and then it prunes entries as needed based on current
        user permissions.

        The web app calls this method but you normally should not need
        to override it; you can override :meth:`make_menus()` instead.
        """
        raw_menus = self._make_raw_menus(request, **kwargs)

        # now we have "simple" (raw) menus definition, but must refine
        # that somewhat to produce our final menus
        self._mark_allowed(request, raw_menus)
        final_menus = []
        for topitem in raw_menus:

            if topitem['allowed']:

                if topitem.get('type') == 'link':
                    final_menus.append(self._make_menu_entry(request, topitem))

                else: # assuming 'menu' type

                    menu_items = []
                    for item in topitem['items']:
                        if not item['allowed']:
                            continue

                        # nested submenu
                        if item.get('type') == 'menu':
                            submenu_items = []
                            for subitem in item['items']:
                                if subitem['allowed']:
                                    submenu_items.append(self._make_menu_entry(request, subitem))
                            menu_items.append({
                                'type': 'submenu',
                                'title': item['title'],
                                'items': submenu_items,
                                'is_menu': True,
                                'is_sep': False,
                            })

                        elif item.get('type') == 'sep':
                            # we only want to add a sep, *if* we already have some
                            # menu items (i.e. there is something to separate)
                            # *and* the last menu item is not a sep (avoid doubles)
                            if menu_items and not menu_items[-1]['is_sep']:
                                menu_items.append(self._make_menu_entry(request, item))

                        else: # standard menu item
                            menu_items.append(self._make_menu_entry(request, item))

                    # remove final separator if present
                    if menu_items and menu_items[-1]['is_sep']:
                        menu_items.pop()

                    # only add if we wound up with something
                    assert menu_items
                    if menu_items:
                        group = {
                            'type': 'menu',
                            'key': topitem.get('key'),
                            'title': topitem['title'],
                            'items': menu_items,
                            'is_menu': True,
                            'is_link':  False,
                        }

                        # topitem w/ no key likely means it did not come
                        # from config but rather explicit definition in
                        # code.  so we are free to "invent" a (safe) key
                        # for it, since that is only for editing config
                        if not group['key']:
                            group['key'] = self._make_menu_key(topitem['title'])

                        final_menus.append(group)

        return final_menus

    def _make_raw_menus(self, request, **kwargs):
        """
        Construct the initial full set of "raw" menus.

        For now this just calls :meth:`make_menus()` which generally
        means a "hard-coded" menu set.  Eventually it may allow for
        loading dynamic menus from config instead.
        """
        return self.make_menus(request, **kwargs)

    def _is_allowed(self, request, item):
        """
        Logic to determine if a given menu item is "allowed" for
        current user.
        """
        perm = item.get('perm')
        if perm:
            return request.has_perm(perm)
        return True

    def _mark_allowed(self, request, menus):
        """
        Traverse the menu set, and mark each item as "allowed" (or
        not) based on current user permissions.
        """
        for topitem in menus:

            if topitem.get('type', 'menu') == 'link':
                topitem['allowed'] = True

            elif topitem.get('type', 'menu') == 'menu':
                topitem['allowed'] = False

                for item in topitem['items']:

                    if item.get('type') == 'menu':
                        for subitem in item['items']:
                            subitem['allowed'] = self._is_allowed(request, subitem)

                        item['allowed'] = False
                        for subitem in item['items']:
                            if subitem['allowed'] and subitem.get('type') != 'sep':
                                item['allowed'] = True
                                break

                    else:
                        item['allowed'] = self._is_allowed(request, item)

                for item in topitem['items']:
                    if item['allowed'] and item.get('type') != 'sep':
                        topitem['allowed'] = True
                        break

    def _make_menu_entry(self, request, item):
        """
        Convert a simple menu entry dict, into a proper menu-related
        object, for use in constructing final menu.
        """
        # separator
        if item.get('type') == 'sep':
            return {
                'type': 'sep',
                'is_menu': False,
                'is_sep': True,
            }

        # standard menu item
        entry = {
            'type': 'item',
            'title': item['title'],
            'perm': item.get('perm'),
            'target': item.get('target'),
            'is_link': True,
            'is_menu': False,
            'is_sep': False,
        }
        if item.get('route'):
            entry['route'] = item['route']
            try:
                entry['url'] = request.route_url(entry['route'])
            except KeyError:        # happens if no such route
                log.warning("invalid route name for menu entry: %s", entry)
                entry['url'] = entry['route']
            entry['key'] = entry['route']
        else:
            if item.get('url'):
                entry['url'] = item['url']
            entry['key'] = self._make_menu_key(entry['title'])
        return entry

    def _make_menu_key(self, value):
        """
        Generate a normalized menu key for the given value.
        """
        return re.sub(r'\W', '', value.lower())



# -*- coding: utf-8; -*-

import datetime
import decimal
from collections import OrderedDict
from enum import Enum
from unittest import TestCase
from unittest.mock import patch, MagicMock

import sqlalchemy as sa
from sqlalchemy import orm
from paginate import Page
from paginate_sqlalchemy import SqlalchemyOrmPage
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttaweb.grids import base as mod
from wuttaweb.grids.filters import GridFilter, StringAlchemyFilter, default_sqlalchemy_filters
from wuttaweb.util import FieldList
from wuttaweb.forms import Form
from wuttaweb.testing import WebTestCase


class TestGrid(WebTestCase):

    def make_grid(self, request=None, **kwargs):
        return mod.Grid(request or self.request, **kwargs)

    def test_constructor(self):

        # empty
        grid = self.make_grid()
        self.assertIsNone(grid.key)
        self.assertEqual(grid.columns, [])
        self.assertIsNone(grid.data)

        # now with columns
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertIsInstance(grid.columns, FieldList)
        self.assertEqual(grid.columns, ['foo', 'bar'])

    def test_constructor_sorting(self):
        model = self.app.model

        # defaults, not sortable
        grid = self.make_grid()
        self.assertFalse(grid.sortable)
        self.assertTrue(grid.sort_on_backend)
        self.assertEqual(grid.sorters, {})
        self.assertEqual(grid.sort_defaults, [])

        # defaults, sortable
        grid = self.make_grid(sortable=True)
        self.assertTrue(grid.sortable)
        self.assertTrue(grid.sort_on_backend)
        self.assertEqual(grid.sorters, {})
        self.assertEqual(grid.sort_defaults, [])

        # sorters may be pre-populated
        grid = self.make_grid(model_class=model.Setting, sortable=True)
        self.assertEqual(len(grid.sorters), 2)
        self.assertIn('name', grid.sorters)
        self.assertIn('value', grid.sorters)
        self.assertEqual(grid.sort_defaults, [])

        # sort defaults as str
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_defaults='name')
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'asc')])

        # sort defaults as tuple
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_defaults=('name', 'desc'))
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc')])

        # sort defaults as list w/ single tuple
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_defaults=[('name', 'desc')])
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc')])

        # multi-column defaults
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_multiple=True,
                              sort_defaults=[('name', 'desc'), ('value', 'asc')])
        self.assertTrue(grid.sort_multiple)
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc'),
                                              mod.SortInfo('value', 'asc')])

        # multi-column sort disabled for oruga
        self.request.use_oruga = True
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_multiple=True)
        self.assertFalse(grid.sort_multiple)

    def test_constructor_filtering(self):
        model = self.app.model

        # defaults, not filterable
        grid = self.make_grid()
        self.assertFalse(grid.filterable)
        self.assertEqual(grid.filters, {})

        # defaults, filterable
        grid = self.make_grid(filterable=True)
        self.assertTrue(grid.filterable)
        self.assertEqual(grid.filters, {})

        # filters may be pre-populated
        with patch.object(mod.Grid, 'make_filter', return_value=42):
            grid = self.make_grid(model_class=model.Setting, filterable=True)
            self.assertEqual(len(grid.filters), 2)
            self.assertIn('name', grid.filters)
            self.assertIn('value', grid.filters)

        # can specify filters
        grid = self.make_grid(model_class=model.Setting, filterable=True,
                              filters={'name': 42})
        self.assertTrue(grid.filterable)
        self.assertEqual(grid.filters, {'name': 42})

    def test_vue_tagname(self):
        grid = self.make_grid()
        self.assertEqual(grid.vue_tagname, 'wutta-grid')

    def test_vue_component(self):
        grid = self.make_grid()
        self.assertEqual(grid.vue_component, 'WuttaGrid')

    def test_get_columns(self):
        model = self.app.model

        # empty
        grid = self.make_grid()
        self.assertEqual(grid.columns, [])
        self.assertEqual(grid.get_columns(), [])

        # explicit
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.columns, ['foo', 'bar'])
        self.assertEqual(grid.get_columns(), ['foo', 'bar'])

        # derived from model
        grid = self.make_grid(model_class=model.Setting)
        self.assertEqual(grid.columns, ['name', 'value'])
        self.assertEqual(grid.get_columns(), ['name', 'value'])

    def test_append(self):
        grid = self.make_grid(columns=['one', 'two'])
        self.assertEqual(grid.columns, ['one', 'two'])
        grid.append('one', 'two', 'three')
        self.assertEqual(grid.columns, ['one', 'two', 'three'])

    def test_remove(self):
        grid = self.make_grid(columns=['one', 'two', 'three', 'four'])
        self.assertEqual(grid.columns, ['one', 'two', 'three', 'four'])
        grid.remove('two', 'three')
        self.assertEqual(grid.columns, ['one', 'four'])

    def test_set_label(self):
        model = self.app.model
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting, filterable=True)
        self.assertEqual(grid.labels, {})

        # basic
        grid.set_label('name', "NAME COL")
        self.assertEqual(grid.labels['name'], "NAME COL")

        # can replace label
        grid.set_label('name', "Different")
        self.assertEqual(grid.labels['name'], "Different")
        self.assertEqual(grid.get_label('name'), "Different")

        # can update only column, not filter
        self.assertEqual(grid.labels, {'name': "Different"})
        self.assertIn('name', grid.filters)
        self.assertEqual(grid.filters['name'].label, "Different")
        grid.set_label('name', "COLUMN ONLY", column_only=True)
        self.assertEqual(grid.get_label('name'), "COLUMN ONLY")
        self.assertEqual(grid.filters['name'].label, "Different")

    def test_get_label(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.labels, {})

        # default derived from key
        self.assertEqual(grid.get_label('foo'), "Foo")

        # can override
        grid.set_label('foo', "Different")
        self.assertEqual(grid.get_label('foo'), "Different")

    def test_set_renderer(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.renderers, {})

        def render1(record, key, value):
            pass

        # basic
        grid.set_renderer('foo', render1)
        self.assertIs(grid.renderers['foo'], render1)

        def render2(record, key, value, extra=None):
            return extra

        # can pass kwargs to get a partial
        grid.set_renderer('foo', render2, extra=42)
        self.assertIsNot(grid.renderers['foo'], render2)
        self.assertEqual(grid.renderers['foo'](None, None, None), 42)

        # can use built-in string shortcut
        grid.set_renderer('foo', 'quantity')
        obj = MagicMock(foo=42.00)
        self.assertEqual(grid.renderers['foo'](obj, 'foo', 42.00), '42')

    def test_set_default_renderers(self):
        model = self.app.model

        # no defaults for "plain" schema
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.renderers, {})

        # no defaults for "plain" mapped class
        grid = self.make_grid(model_class=model.Setting)
        self.assertEqual(grid.renderers, {})

        def myrender(obj, key, value):
            return value

        # renderer set for datetime mapped field
        grid = self.make_grid(model_class=model.Upgrade)
        self.assertIn('created', grid.renderers)
        self.assertIsNot(grid.renderers['created'], myrender)

        # renderer *not* set for datetime, if override present
        grid = self.make_grid(model_class=model.Upgrade,
                              renderers={'created': myrender})
        self.assertIn('created', grid.renderers)
        self.assertIs(grid.renderers['created'], myrender)

        # renderer set for boolean mapped field
        grid = self.make_grid(model_class=model.Upgrade)
        self.assertIn('executing', grid.renderers)
        self.assertIsNot(grid.renderers['executing'], myrender)

        # renderer *not* set for boolean, if override present
        grid = self.make_grid(model_class=model.Upgrade,
                              renderers={'executing': myrender})
        self.assertIn('executing', grid.renderers)
        self.assertIs(grid.renderers['executing'], myrender)

        # nb. as of writing we have no Date columns in default schema,
        # so must invent one to test that type
        class SomeFoolery(model.Base):
            __tablename__ = 'somefoolery'
            id = sa.Column(sa.Integer(), primary_key=True)
            created = sa.Column(sa.Date())

        # renderer set for date mapped field
        grid = self.make_grid(model_class=SomeFoolery)
        self.assertIn('created', grid.renderers)
        self.assertIsNot(grid.renderers['created'], myrender)

    def test_set_enum(self):
        model = self.app.model

        class MockEnum(Enum):
            FOO = 'foo'
            BAR = 'bar'

        # no enums by default
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.enums, {})

        # enum is set, but not filter choices
        grid = self.make_grid(columns=['foo', 'bar'],
                              filterable=False,
                              enums={'foo': MockEnum})
        self.assertIs(grid.enums['foo'], MockEnum)
        self.assertEqual(grid.filters, {})

        # both enum and filter choices are set
        grid = self.make_grid(model_class=model.Setting,
                              filterable=True,
                              enums={'name': MockEnum})
        self.assertIs(grid.enums['name'], MockEnum)
        self.assertIn('name', grid.filters)
        self.assertIn('value', grid.filters)
        self.assertEqual(grid.filters['name'].choices, OrderedDict([
            ('FOO', 'foo'),
            ('BAR', 'bar'),
        ]))

    def test_linked_columns(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.linked_columns, [])
        self.assertFalse(grid.is_linked('foo'))

        grid.set_link('foo')
        self.assertEqual(grid.linked_columns, ['foo'])
        self.assertTrue(grid.is_linked('foo'))
        self.assertFalse(grid.is_linked('bar'))

        grid.set_link('bar')
        self.assertEqual(grid.linked_columns, ['foo', 'bar'])
        self.assertTrue(grid.is_linked('foo'))
        self.assertTrue(grid.is_linked('bar'))

        grid.set_link('foo', False)
        self.assertEqual(grid.linked_columns, ['bar'])
        self.assertFalse(grid.is_linked('foo'))
        self.assertTrue(grid.is_linked('bar'))

    def test_hidden_columns(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.hidden_columns, [])
        self.assertFalse(grid.is_hidden('foo'))

        grid.set_hidden('foo')
        self.assertEqual(grid.hidden_columns, ['foo'])
        self.assertTrue(grid.is_hidden('foo'))
        self.assertFalse(grid.is_hidden('bar'))

        grid.set_hidden('bar')
        self.assertEqual(grid.hidden_columns, ['foo', 'bar'])
        self.assertTrue(grid.is_hidden('foo'))
        self.assertTrue(grid.is_hidden('bar'))

        grid.set_hidden('foo', False)
        self.assertEqual(grid.hidden_columns, ['bar'])
        self.assertFalse(grid.is_hidden('foo'))
        self.assertTrue(grid.is_hidden('bar'))

    def test_searchable_columns(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.searchable_columns, set())
        self.assertFalse(grid.is_searchable('foo'))

        grid.set_searchable('foo')
        self.assertEqual(grid.searchable_columns, {'foo'})
        self.assertTrue(grid.is_searchable('foo'))
        self.assertFalse(grid.is_searchable('bar'))

        grid.set_searchable('bar')
        self.assertEqual(grid.searchable_columns, {'foo', 'bar'})
        self.assertTrue(grid.is_searchable('foo'))
        self.assertTrue(grid.is_searchable('bar'))

        grid.set_searchable('foo', False)
        self.assertEqual(grid.searchable_columns, {'bar'})
        self.assertFalse(grid.is_searchable('foo'))
        self.assertTrue(grid.is_searchable('bar'))

    def test_add_action(self):
        grid = self.make_grid()
        self.assertEqual(len(grid.actions), 0)

        grid.add_action('view')
        self.assertEqual(len(grid.actions), 1)
        self.assertIsInstance(grid.actions[0], mod.GridAction)

    def test_set_tools(self):
        grid = self.make_grid()
        self.assertEqual(grid.tools, {})

        # null
        grid.set_tools(None)
        self.assertEqual(grid.tools, {})

        # empty
        grid.set_tools({})
        self.assertEqual(grid.tools, {})

        # full dict is replaced
        grid.tools = {'foo': 'bar'}
        self.assertEqual(grid.tools, {'foo': 'bar'})
        grid.set_tools({'bar': 'baz'})
        self.assertEqual(grid.tools, {'bar': 'baz'})

        # can specify as list of html elements
        grid.set_tools(['foo', 'bar'])
        self.assertEqual(len(grid.tools), 2)
        self.assertEqual(list(grid.tools.values()), ['foo', 'bar'])

    def test_add_tool(self):
        grid = self.make_grid()
        self.assertEqual(grid.tools, {})

        # with key
        grid.add_tool('foo', key='foo')
        self.assertEqual(grid.tools, {'foo': 'foo'})

        # without key
        grid.add_tool('bar')
        self.assertEqual(len(grid.tools), 2)
        self.assertEqual(list(grid.tools.values()), ['foo', 'bar'])

    def test_get_pagesize_options(self):
        grid = self.make_grid()

        # default
        options = grid.get_pagesize_options()
        self.assertEqual(options, [5, 10, 20, 50, 100, 200])

        # override default
        options = grid.get_pagesize_options(default=[42])
        self.assertEqual(options, [42])

        # from config
        self.config.setdefault('wuttaweb.grids.default_pagesize_options', '1 2 3')
        options = grid.get_pagesize_options()
        self.assertEqual(options, [1, 2, 3])

    def test_get_pagesize(self):
        grid = self.make_grid()

        # default
        size = grid.get_pagesize()
        self.assertEqual(size, 20)

        # override default
        size = grid.get_pagesize(default=42)
        self.assertEqual(size, 42)

        # override default options
        self.config.setdefault('wuttaweb.grids.default_pagesize_options', '10 15 30')
        grid = self.make_grid()
        size = grid.get_pagesize()
        self.assertEqual(size, 10)

        # from config
        self.config.setdefault('wuttaweb.grids.default_pagesize', '15')
        size = grid.get_pagesize()
        self.assertEqual(size, 15)

    ##############################
    # configuration methods
    ##############################

    def test_load_settings(self):
        model = self.app.model

        # nb. first use a paging grid
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True,
                              pagesize=20, page=1)

        # settings are loaded, applied, saved
        self.assertEqual(grid.page, 1)
        self.assertNotIn('grid.foo.page', self.request.session)
        self.request.GET = {'pagesize': '10', 'page': '2'}
        grid.load_settings()
        self.assertEqual(grid.page, 2)
        self.assertEqual(self.request.session['grid.foo.page'], 2)

        # can skip the saving step
        self.request.GET = {'pagesize': '10', 'page': '3'}
        grid.load_settings(persist=False)
        self.assertEqual(grid.page, 3)
        self.assertEqual(self.request.session['grid.foo.page'], 2)

        # no error for non-paginated grid
        grid = self.make_grid(key='foo', paginated=False)
        grid.load_settings()
        self.assertFalse(grid.paginated)

        # nb. next use a sorting grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True)

        # settings are loaded, applied, saved
        self.assertEqual(grid.sort_defaults, [])
        self.assertFalse(hasattr(grid, 'active_sorters'))
        self.request.GET = {'sort1key': 'name', 'sort1dir': 'desc'}
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'desc'}])
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')

        # can skip the saving step
        self.request.GET = {'sort1key': 'name', 'sort1dir': 'asc'}
        grid.load_settings(persist=False)
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')

        # no error for non-sortable grid
        grid = self.make_grid(key='foo', sortable=False)
        grid.load_settings()
        self.assertFalse(grid.sortable)

        # with sort defaults
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_on_backend=True, sort_defaults='name')
        self.assertFalse(hasattr(grid, 'active_sorters'))
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])

        # with multi-column sort defaults
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_on_backend=True)
        grid.sort_defaults = [
            mod.SortInfo('name', 'asc'),
            mod.SortInfo('value', 'desc'),
        ]
        self.assertFalse(hasattr(grid, 'active_sorters'))
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])

        # load settings from session when nothing is in request
        self.request.GET = {}
        self.request.session.invalidate()
        self.assertNotIn('grid.settings.sorters.length', self.request.session)
        self.request.session['grid.settings.sorters.length'] = 1
        self.request.session['grid.settings.sorters.1.key'] = 'name'
        self.request.session['grid.settings.sorters.1.dir'] = 'desc'
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True,
                              paginated=True, paginate_on_backend=True)
        self.assertFalse(hasattr(grid, 'active_sorters'))
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'desc'}])

        # filter settings are loaded, applied, saved
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True)
        self.assertEqual(len(grid.filters), 2)
        self.assertEqual(len(grid.active_filters), 0)
        self.assertNotIn('grid.settings.filter.name.active', self.request.session)
        self.assertNotIn('grid.settings.filter.value.active', self.request.session)
        self.request.GET = {'name': 'john', 'name.verb': 'contains'}
        grid.load_settings()
        self.assertTrue(grid.filters['name'].active)
        self.assertEqual(grid.filters['name'].verb, 'contains')
        self.assertEqual(grid.filters['name'].value, 'john')
        self.assertTrue(self.request.session['grid.settings.filter.name.active'])
        self.assertEqual(self.request.session['grid.settings.filter.name.verb'], 'contains')
        self.assertEqual(self.request.session['grid.settings.filter.name.value'], 'john')

        # filter + sort settings are loaded, applied, saved
        self.request.session.invalidate()
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, filterable=True)
        self.assertEqual(len(grid.filters), 2)
        self.assertEqual(len(grid.active_filters), 0)
        self.assertNotIn('grid.settings.filter.name.active', self.request.session)
        self.assertNotIn('grid.settings.filter.value.active', self.request.session)
        self.assertNotIn('grid.settings.sorters.length', self.request.session)
        self.request.GET = {'name': 'john', 'name.verb': 'contains',
                            'sort1key': 'name', 'sort1dir': 'asc'}
        grid.load_settings()
        self.assertTrue(grid.filters['name'].active)
        self.assertEqual(grid.filters['name'].verb, 'contains')
        self.assertEqual(grid.filters['name'].value, 'john')
        self.assertTrue(self.request.session['grid.settings.filter.name.active'])
        self.assertEqual(self.request.session['grid.settings.filter.name.verb'], 'contains')
        self.assertEqual(self.request.session['grid.settings.filter.name.value'], 'john')
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'asc')

        # can reset view to defaults
        self.request.GET = {'reset-view': 'true'}
        grid.load_settings()
        self.assertEqual(grid.active_filters, [])
        self.assertIsNone(grid.filters['name'].value)

    def test_request_has_settings(self):
        model = self.app.model
        grid = self.make_grid(key='settings', model_class=model.Setting)

        # paging
        self.assertFalse(grid.request_has_settings('page'))
        with patch.object(grid, 'paginated', new=True):
            with patch.object(self.request, 'GET', new={'pagesize': '20'}):
                self.assertTrue(grid.request_has_settings('page'))
            with patch.object(self.request, 'GET', new={'page': '1'}):
                self.assertTrue(grid.request_has_settings('page'))

        # sorting
        self.assertFalse(grid.request_has_settings('sort'))
        with patch.object(grid, 'sortable', new=True):
            with patch.object(self.request, 'GET', new={'sort1key': 'name'}):
                self.assertTrue(grid.request_has_settings('sort'))

        # filtering
        grid = self.make_grid(key='settings', model_class=model.Setting, filterable=True)
        self.assertFalse(grid.request_has_settings('filter'))
        with patch.object(grid, 'filterable', new=True):
            with patch.object(self.request, 'GET', new={'name': 'john', 'name.verb': 'contains'}):
                self.assertTrue(grid.request_has_settings('filter'))
        with patch.object(self.request, 'GET', new={'filter': '1'}):
            self.assertTrue(grid.request_has_settings('filter'))

    def test_get_setting(self):
        grid = self.make_grid(key='foo')
        settings = {}

        # default is null
        value = grid.get_setting(settings, 'pagesize')
        self.assertIsNone(value)

        # can read value from user session
        self.request.session['grid.foo.pagesize'] = 15
        value = grid.get_setting(settings, 'pagesize', src='session')
        self.assertEqual(value, 15)

        # string value not normalized
        self.request.session['grid.foo.pagesize'] = '15'
        value = grid.get_setting(settings, 'pagesize', src='session')
        self.assertEqual(value, '15')
        self.assertNotEqual(value, 15)

        # but can be normalized
        self.request.session['grid.foo.pagesize'] = '15'
        value = grid.get_setting(settings, 'pagesize', src='session', normalize=int)
        self.assertEqual(value, 15)

        # can read value from request
        self.request.GET = {'pagesize': '25'}
        value = grid.get_setting(settings, 'pagesize', src='request', normalize=int)
        self.assertEqual(value, 25)

        # null when normalization fails
        self.request.GET = {'pagesize': 'invalid'}
        value = grid.get_setting(settings, 'pagesize', src='request', normalize=int)
        self.assertIsNone(value)

        # reset
        del self.request.session['grid.foo.pagesize']
        self.request.GET = {}

        # value can come from provided settings
        settings['pagesize'] = '35'
        value = grid.get_setting(settings, 'pagesize', src='session', normalize=int)
        self.assertEqual(value, 35)

    def test_update_filter_settings(self):
        model = self.app.model

        # nothing happens if not filterable
        grid = self.make_grid(key='settings', model_class=model.Setting)
        settings = {}
        self.request.session['grid.settings.filter.name.active'] = True
        self.request.session['grid.settings.filter.name.verb'] = 'contains'
        self.request.session['grid.settings.filter.name.value'] = 'john'
        grid.update_filter_settings(settings, src='session')
        self.assertEqual(settings, {})

        # nb. now use a filterable grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True)

        # settings are updated from session
        settings = {}
        self.request.session['grid.settings.filter.name.active'] = True
        self.request.session['grid.settings.filter.name.verb'] = 'contains'
        self.request.session['grid.settings.filter.name.value'] = 'john'
        grid.update_filter_settings(settings, src='session')
        self.assertTrue(settings['filter.name.active'])
        self.assertEqual(settings['filter.name.verb'], 'contains')
        self.assertEqual(settings['filter.name.value'], 'john')

        # settings are updated from request
        self.request.GET = {'value': 'sally', 'value.verb': 'contains'}
        grid.update_filter_settings(settings, src='request')
        self.assertFalse(settings['filter.name.active'])
        self.assertTrue(settings['filter.value.active'])
        self.assertEqual(settings['filter.value.verb'], 'contains')
        self.assertEqual(settings['filter.value.value'], 'sally')

    def test_update_sort_settings(self):
        model = self.app.model

        # nothing happens if not sortable
        grid = self.make_grid(key='settings', model_class=model.Setting)
        settings = {'sorters.length': 0}
        self.request.session['grid.settings.sorters.length'] = 1
        self.request.session['grid.settings.sorters.1.key'] = 'name'
        self.request.session['grid.settings.sorters.1.dir'] = 'asc'
        grid.update_sort_settings(settings, src='session')
        self.assertEqual(settings['sorters.length'], 0)

        # nb. now use a sortable grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True)

        # settings are updated from session
        settings = {'sorters.length': 1, 'sorters.1.key': 'name', 'sorters.1.dir': 'asc'}
        self.request.session['grid.settings.sorters.length'] = 1
        self.request.session['grid.settings.sorters.1.key'] = 'name'
        self.request.session['grid.settings.sorters.1.dir'] = 'asc'
        grid.update_sort_settings(settings, src='session')
        self.assertEqual(settings['sorters.length'], 1)
        self.assertEqual(settings['sorters.1.key'], 'name')
        self.assertEqual(settings['sorters.1.dir'], 'asc')

        # settings are updated from request
        self.request.GET = {'sort1key': 'value', 'sort1dir': 'desc'}
        grid.update_sort_settings(settings, src='request')
        self.assertEqual(settings['sorters.length'], 1)
        self.assertEqual(settings['sorters.1.key'], 'value')
        self.assertEqual(settings['sorters.1.dir'], 'desc')

    def test_update_page_settings(self):

        # nothing happens if not paginated
        grid = self.make_grid(key='foo')
        settings = {'pagesize': 20, 'page': 1}
        self.request.session['grid.foo.pagesize'] = 10
        self.request.session['grid.foo.page'] = 2
        grid.update_page_settings(settings)
        self.assertEqual(settings['pagesize'], 20)
        self.assertEqual(settings['page'], 1)

        # nb. now use a paginated grid
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True)

        # settings are updated from session
        settings = {'pagesize': 20, 'page': 1}
        self.request.session['grid.foo.pagesize'] = 10
        self.request.session['grid.foo.page'] = 2
        grid.update_page_settings(settings)
        self.assertEqual(settings['pagesize'], 10)
        self.assertEqual(settings['page'], 2)

        # settings are updated from request
        self.request.GET = {'pagesize': '15', 'page': '4'}
        grid.update_page_settings(settings)
        self.assertEqual(settings['pagesize'], 15)
        self.assertEqual(settings['page'], 4)

    def test_persist_settings(self):
        model = self.app.model

        # nb. start out with paginated-only grid
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True)

        # invalid dest
        self.assertRaises(ValueError, grid.persist_settings, {}, dest='doesnotexist')

        # nb. no error if empty settings, but it saves null values
        grid.persist_settings({}, dest='session')
        self.assertIsNone(self.request.session['grid.foo.page'])

        # provided values are saved
        grid.persist_settings({'pagesize': 15, 'page': 3}, dest='session')
        self.assertEqual(self.request.session['grid.foo.page'], 3)

        # nb. now switch to sortable-only grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True)

        # no error if empty settings; does not save values
        grid.persist_settings({}, dest='session')
        self.assertNotIn('grid.settings.sorters.length', self.request.session)

        # provided values are saved
        grid.persist_settings({'sorters.length': 2,
                               'sorters.1.key': 'name',
                               'sorters.1.dir': 'desc',
                               'sorters.2.key': 'value',
                               'sorters.2.dir': 'asc'},
                              dest='session')
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 2)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')
        self.assertEqual(self.request.session['grid.settings.sorters.2.key'], 'value')
        self.assertEqual(self.request.session['grid.settings.sorters.2.dir'], 'asc')

        # old values removed when new are saved
        grid.persist_settings({'sorters.length': 1,
                               'sorters.1.key': 'name',
                               'sorters.1.dir': 'desc'},
                              dest='session')
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')
        self.assertNotIn('grid.settings.sorters.2.key', self.request.session)
        self.assertNotIn('grid.settings.sorters.2.dir', self.request.session)

        # nb. now switch to filterable-only grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True)
        self.assertIn('name', grid.filters)
        self.assertEqual(grid.filters['name'].key, 'name')

        # no error if empty settings; does not save values
        grid.persist_settings({}, dest='session')
        self.assertNotIn('grid.settings.filters.name', self.request.session)

        # provided values are saved
        grid.persist_settings({'filter.name.active': True,
                               'filter.name.verb': 'contains',
                               'filter.name.value': 'john'},
                              dest='session')
        self.assertTrue(self.request.session['grid.settings.filter.name.active'])
        self.assertEqual(self.request.session['grid.settings.filter.name.verb'], 'contains')
        self.assertEqual(self.request.session['grid.settings.filter.name.value'], 'john')

    ##############################
    # sorting methods
    ##############################

    def test_make_backend_sorters(self):
        model = self.app.model

        # default is empty
        grid = self.make_grid()
        sorters = grid.make_backend_sorters()
        self.assertEqual(sorters, {})

        # makes sorters if model class
        grid = self.make_grid(model_class=model.Setting)
        sorters = grid.make_backend_sorters()
        self.assertEqual(len(sorters), 2)
        self.assertIn('name', sorters)
        self.assertIn('value', sorters)

        # does not replace supplied sorters
        grid = self.make_grid(model_class=model.Setting)
        mysorters = {'value': 42}
        sorters = grid.make_backend_sorters(mysorters)
        self.assertEqual(len(sorters), 2)
        self.assertIn('name', sorters)
        self.assertIn('value', sorters)
        self.assertEqual(sorters['value'], 42)
        self.assertEqual(mysorters['value'], 42)

    def test_make_sorter(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        # plain data
        grid = self.make_grid(columns=['name', 'value'])
        sorter = grid.make_sorter('name')
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(sorted_data[0], {'name': 'foo9', 'value': 'nine'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(sorted_data[0], {'name': 'foo1', 'value': 'ONE'})

        # model class, but still plain data
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('name')
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(sorted_data[0], {'name': 'foo9', 'value': 'nine'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(sorted_data[0], {'name': 'foo1', 'value': 'ONE'})

        # repeat previous test, w/ model property
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter(model.Setting.name)
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(sorted_data[0], {'name': 'foo9', 'value': 'nine'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(sorted_data[0], {'name': 'foo1', 'value': 'ONE'})

        # sqlalchemy query
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('name')
        sorted_query = sorter(sample_query, 'desc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo9', 'value': 'nine'})
        sorted_query = sorter(sample_query, 'asc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo1', 'value': 'ONE'})

        # repeat previous test, w/ model property
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter(model.Setting.name)
        sorted_query = sorter(sample_query, 'desc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo9', 'value': 'nine'})
        sorted_query = sorter(sample_query, 'asc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo1', 'value': 'ONE'})

        # sortfunc for "invalid" column will fail when called; however
        # it can work for manual sort w/ custom keyfunc
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('doesnotexist')
        self.assertRaises(TypeError, sorter, sample_query, 'desc')
        self.assertRaises(KeyError, sorter, sample_data, 'desc')
        sorter = grid.make_sorter('doesnotexist', keyfunc=lambda obj: obj['name'])
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(len(sorted_data), 9)
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(len(sorted_data), 9)

        # case folding is on by default
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('value')
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo2', 'value': 'two'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo8', 'value': 'eight'})

        # results are different with case folding off
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('value', foldcase=False)
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo2', 'value': 'two'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo1', 'value': 'ONE'})

    def test_set_joiner(self):

        # basic
        grid = self.make_grid(columns=['foo', 'bar'], sortable=True, sort_on_backend=True)
        self.assertEqual(grid.joiners, {})
        grid.set_joiner('foo', 42)
        self.assertEqual(grid.joiners, {'foo': 42})

    def test_remove_joiner(self):

        # basic
        grid = self.make_grid(columns=['foo', 'bar'], sortable=True, sort_on_backend=True,
                              joiners={'foo': 42})
        self.assertEqual(grid.joiners, {'foo': 42})
        grid.remove_joiner('foo')
        self.assertEqual(grid.joiners, {})

    def test_set_sorter(self):
        model = self.app.model

        # explicit sortfunc
        grid = self.make_grid()
        self.assertEqual(grid.sorters, {})
        sortfunc = lambda data, direction: data
        grid.set_sorter('foo', sortfunc)
        self.assertIs(grid.sorters['foo'], sortfunc)

        # auto from model property
        grid = self.make_grid(model_class=model.Setting, sortable=True, sorters={})
        self.assertEqual(grid.sorters, {})
        grid.set_sorter('name', model.Setting.name)
        self.assertTrue(callable(grid.sorters['name']))

        # auto from column name
        grid = self.make_grid(model_class=model.Setting, sortable=True, sorters={})
        self.assertEqual(grid.sorters, {})
        grid.set_sorter('name', 'name')
        self.assertTrue(callable(grid.sorters['name']))

        # auto from key
        grid = self.make_grid(model_class=model.Setting, sortable=True, sorters={})
        self.assertEqual(grid.sorters, {})
        grid.set_sorter('name')
        self.assertTrue(callable(grid.sorters['name']))

    def test_remove_sorter(self):
        model = self.app.model

        # basics
        grid = self.make_grid(model_class=model.Setting, sortable=True)
        self.assertEqual(len(grid.sorters), 2)
        self.assertIn('name', grid.sorters)
        self.assertIn('value', grid.sorters)
        grid.remove_sorter('value')
        self.assertNotIn('value', grid.sorters)

    def test_set_sort_defaults(self):
        model = self.app.model
        grid = self.make_grid(model_class=model.Setting, sortable=True)
        self.assertEqual(grid.sort_defaults, [])

        # can set just sortkey
        grid.set_sort_defaults('name')
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'asc')])

        # can set sortkey, sortdir
        grid.set_sort_defaults('name', 'desc')
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc')])

        # can set sortkey, sortdir as tuple
        grid.set_sort_defaults(('value', 'asc'))
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('value', 'asc')])

        # can set as list
        grid.sort_multiple = True
        grid.set_sort_defaults([('value', 'asc'), ('name', 'desc')])
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('value', 'asc'),
                                              mod.SortInfo('name', 'desc')])

        # list is pruned if multi-sort disabled
        grid.sort_multiple = False
        grid.set_sort_defaults([('value', 'asc'), ('name', 'desc')])
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('value', 'asc')])

        # error if any other single arg
        self.assertRaises(ValueError, grid.set_sort_defaults, 42)

        # error if more than 2 args
        self.assertRaises(ValueError, grid.set_sort_defaults, 'name', 'asc', 'value', 'desc')

    def test_is_sortable(self):
        model = self.app.model

        # basics, frontend sorting
        grid = self.make_grid(model_class=model.Setting, sortable=True, sort_on_backend=False)
        self.assertTrue(grid.is_sortable('name'))
        self.assertTrue(grid.is_sortable('value'))
        grid.remove_sorter('value')
        # nb. columns are always sortable for frontend, despite remove_sorter()
        self.assertTrue(grid.is_sortable('value'))
        # nb. when grid is not sortable, no column is either
        grid.sortable = False
        self.assertFalse(grid.is_sortable('name'))

        # same test but with backend sorting
        grid = self.make_grid(model_class=model.Setting, sortable=True, sort_on_backend=True)
        self.assertTrue(grid.is_sortable('name'))
        self.assertTrue(grid.is_sortable('value'))
        grid.remove_sorter('value')
        self.assertFalse(grid.is_sortable('value'))
        # nb. when grid is not sortable, no column is either
        grid.sortable = False
        self.assertFalse(grid.is_sortable('name'))

    def test_make_backend_filters(self):
        model = self.app.model

        # default is empty
        grid = self.make_grid()
        filters = grid.make_backend_filters()
        self.assertEqual(filters, {})

        # makes filters if model class
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting)
            filters = grid.make_backend_filters()
        self.assertEqual(len(filters), 2)
        self.assertIn('name', filters)
        self.assertIn('value', filters)

        # does not replace supplied filters
        myfilters = {'value': 42}
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting)
            filters = grid.make_backend_filters(myfilters)
        self.assertEqual(len(filters), 2)
        self.assertIn('name', filters)
        self.assertIn('value', filters)
        self.assertEqual(filters['value'], 42)
        self.assertEqual(myfilters['value'], 42)

        # filters for all *true* columns by default, despite grid.columns
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.User,
                                  columns=['username', 'person'])
            filters = grid.make_backend_filters()
            self.assertIn('username', filters)
            self.assertIn('active', filters)
            # nb. relationship not included by default
            self.assertNotIn('person', filters)
            # nb. uuid fields not included by default
            self.assertNotIn('uuid', filters)
            self.assertNotIn('person_uuid', filters)

    def test_make_filter(self):
        model = self.app.model

        # arg is column name
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter('name')
        self.assertIsInstance(filtr, StringAlchemyFilter)

        # arg is column name, but model class is invalid
        grid = self.make_grid(model_class=42)
        self.assertRaises(ValueError, grid.make_filter, 'name')

        # arg is model property
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter(model.Setting.name)
        self.assertIsInstance(filtr, StringAlchemyFilter)

        # model property as kwarg
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter(None, model_property=model.Setting.name)
        self.assertIsInstance(filtr, StringAlchemyFilter)

        # default factory
        grid = self.make_grid(model_class=model.Setting)
        with patch.dict(default_sqlalchemy_filters, {None: GridFilter}, clear=True):
            filtr = grid.make_filter(model.Setting.name)
        self.assertIsInstance(filtr, GridFilter)
        self.assertNotIsInstance(filtr, StringAlchemyFilter)

        # factory override
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter(model.Setting.name, factory=GridFilter)
        self.assertIsInstance(filtr, GridFilter)
        self.assertNotIsInstance(filtr, StringAlchemyFilter)

    def test_set_filter(self):
        model = self.app.model

        with patch.object(mod.Grid, 'make_filter', return_value=42):

            # auto from model property
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            grid.set_filter('name', model.Setting.name)
            self.assertIn('name', grid.filters)

            # auto from column name
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            grid.set_filter('name', 'name')
            self.assertIn('name', grid.filters)

            # auto from key
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            grid.set_filter('name')
            self.assertIn('name', grid.filters)

            # explicit is not yet implemented
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            self.assertRaises(NotImplementedError, grid.set_filter, 'name', lambda q: q)

    def test_remove_filter(self):
        model = self.app.model

        # basics
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting, filterable=True)
        self.assertEqual(len(grid.filters), 2)
        self.assertIn('name', grid.filters)
        self.assertIn('value', grid.filters)
        grid.remove_filter('value')
        self.assertNotIn('value', grid.filters)

    def test_set_filter_defaults(self):
        model = self.app.model

        # empty by default
        grid = self.make_grid(model_class=model.Setting, filterable=True)
        self.assertEqual(grid.filter_defaults, {})

        # can specify via method call
        grid.set_filter_defaults(name={'active': True})
        self.assertEqual(grid.filter_defaults, {'name': {'active': True}})

        # can specify via constructor
        grid = self.make_grid(model_class=model.Setting, filterable=True,
                              filter_defaults={'name': {'active': True}})
        self.assertEqual(grid.filter_defaults, {'name': {'active': True}})

    ##############################
    # data methods
    ##############################

    def test_get_visible_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        # data is sorted and paginated
        grid = self.make_grid(model_class=model.Setting,
                              data=sample_query,
                              filterable=True,
                              sortable=True, sort_on_backend=True,
                              sort_defaults=('name', 'desc'),
                              paginated=True, paginate_on_backend=True,
                              pagesize=4, page=2)
        grid.load_settings()
        # nb. for now the filtering is mocked
        with patch.object(grid, 'filter_data') as filter_data:
            filter_data.side_effect = lambda q: q
            visible = grid.get_visible_data()
            filter_data.assert_called_once_with(sample_query)
        self.assertEqual([s.name for s in visible], ['foo5', 'foo4', 'foo3', 'foo2'])

    def test_filter_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        grid = self.make_grid(key='settings', model_class=model.Setting, filterable=True)
        self.assertEqual(list(grid.filters), ['name', 'value'])
        self.assertIsInstance(grid.filters['name'], StringAlchemyFilter)
        self.assertIsInstance(grid.filters['value'], StringAlchemyFilter)

        # not filtered by default
        grid.load_settings()
        self.assertEqual(grid.active_filters, [])
        filtered_query = grid.filter_data(sample_query)
        self.assertIs(filtered_query, sample_query)

        # can be filtered per session settings
        self.request.session['grid.settings.filter.value.active'] = True
        self.request.session['grid.settings.filter.value.verb'] = 'contains'
        self.request.session['grid.settings.filter.value.value'] = 'ggg'
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].key, 'value')
        filtered_query = grid.filter_data(sample_query)
        self.assertIsInstance(filtered_query, orm.Query)
        self.assertIsNot(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 3)

        # can be filtered per request settings
        self.request.GET = {'value': 's', 'value.verb': 'contains'}
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].key, 'value')
        filtered_query = grid.filter_data(sample_query)
        self.assertIsInstance(filtered_query, orm.Query)
        self.assertEqual(filtered_query.count(), 2)

        # not filtered if verb is invalid
        self.request.GET = {'value': 'ggg', 'value.verb': 'doesnotexist'}
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].verb, 'doesnotexist')
        filtered_query = grid.filter_data(sample_query)
        self.assertIs(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # not filtered if error
        self.request.GET = {'value': 'ggg', 'value.verb': 'contains'}
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].verb, 'contains')
        filtered_query = grid.filter_data(sample_query)
        self.assertIsNot(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 3)
        with patch.object(grid.active_filters[0], 'filter_contains', side_effect=RuntimeError):
            filtered_query = grid.filter_data(sample_query)
        self.assertIs(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # joiner is invoked
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].key, 'value')
        joiner = MagicMock(side_effect=lambda q: q)
        grid.joiners = {'value': joiner}
        grid.joined = set()
        filtered_query = grid.filter_data(sample_query)
        joiner.assert_called_once_with(sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_sort_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        grid = self.make_grid(model_class=model.Setting,
                              sortable=True, sort_on_backend=True,
                              sort_defaults=('name', 'desc'))
        grid.load_settings()

        # can sort a simple list of data
        sorted_data = grid.sort_data(sample_data)
        self.assertIsInstance(sorted_data, list)
        self.assertEqual(len(sorted_data), 9)
        self.assertEqual(sorted_data[0]['name'], 'foo9')
        self.assertEqual(sorted_data[-1]['name'], 'foo1')

        # can also sort a data query
        sorted_query = grid.sort_data(sample_query)
        self.assertIsInstance(sorted_query, orm.Query)
        sorted_data = sorted_query.all()
        self.assertEqual(len(sorted_data), 9)
        self.assertEqual(sorted_data[0]['name'], 'foo9')
        self.assertEqual(sorted_data[-1]['name'], 'foo1')

        # cannot sort data if sorter missing in overrides
        sorted_data = grid.sort_data(sample_data, sorters=[])
        # nb. sorted data is in same order as original sample (not sorted)
        self.assertEqual(sorted_data[0]['name'], 'foo1')
        self.assertEqual(sorted_data[-1]['name'], 'foo9')

        # multi-column sorting for list data
        sorted_data = grid.sort_data(sample_data, sorters=[{'key': 'value', 'dir': 'asc'},
                                                           {'key': 'name', 'dir': 'asc'}])
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo8', 'value': 'eight'})
        self.assertEqual(dict(sorted_data[1]), {'name': 'foo3', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[3]), {'name': 'foo5', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[-1]), {'name': 'foo2', 'value': 'two'})

        # multi-column sorting for query
        sorted_query = grid.sort_data(sample_query, sorters=[{'key': 'value', 'dir': 'asc'},
                                                             {'key': 'name', 'dir': 'asc'}])
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo8', 'value': 'eight'})
        self.assertEqual(dict(sorted_data[1]), {'name': 'foo3', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[3]), {'name': 'foo5', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[-1]), {'name': 'foo2', 'value': 'two'})

        # cannot sort data if sortfunc is missing for column
        grid.remove_sorter('name')
        sorted_data = grid.sort_data(sample_data, sorters=[{'key': 'value', 'dir': 'asc'},
                                                           {'key': 'name', 'dir': 'asc'}])
        # nb. sorted data is in same order as original sample (not sorted)
        self.assertEqual(sorted_data[0]['name'], 'foo1')
        self.assertEqual(sorted_data[-1]['name'], 'foo9')

        # now try with a joiner
        query = self.session.query(model.User)
        grid = self.make_grid(model_class=model.User,
                              data=query,
                              columns=['username', 'full_name'],
                              sortable=True, sort_on_backend=True,
                              sort_defaults='full_name',
                              joiners={
                                  'full_name': lambda q: q.join(model.Person),
                              })
        grid.set_sorter('full_name', model.Person.full_name)
        grid.load_settings()
        data = grid.get_visible_data()
        self.assertIsInstance(data, orm.Query)

    def test_paginate_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        # basic list pager
        grid = self.make_grid(paginated=True, paginate_on_backend=True)
        pager = grid.paginate_data(sample_data)
        self.assertIsInstance(pager, Page)

        # basic query pager
        grid = self.make_grid(paginated=True, paginate_on_backend=True)
        pager = grid.paginate_data(sample_query)
        self.assertIsInstance(pager, SqlalchemyOrmPage)

        # page is reset to 1 for empty data
        self.request.session['grid.foo.page'] = 2
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True)
        grid.load_settings()
        self.assertEqual(grid.page, 2)
        self.assertEqual(self.request.session['grid.foo.page'], 2)
        pager = grid.paginate_data(sample_data)
        self.assertEqual(pager.page, 1)
        self.assertEqual(grid.page, 1)
        self.assertEqual(self.request.session['grid.foo.page'], 1)

    ##############################
    # rendering methods
    ##############################

    def test_render_batch_id(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        # null
        obj = MagicMock(foo=None)
        self.assertEqual(grid.render_batch_id(obj, 'foo', None), "")

        # int
        obj = MagicMock(foo=42)
        self.assertEqual(grid.render_batch_id(obj, 'foo', 42), "00000042")

    def test_render_boolean(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        # null
        obj = MagicMock(foo=None)
        self.assertEqual(grid.render_boolean(obj, 'foo', None), "")

        # true
        obj = MagicMock(foo=True)
        self.assertEqual(grid.render_boolean(obj, 'foo', True), "Yes")

        # false
        obj = MagicMock(foo=False)
        self.assertEqual(grid.render_boolean(obj, 'foo', False), "No")

    def test_render_currency(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        obj = MagicMock()

        # null
        self.assertEqual(grid.render_currency(obj, 'foo', None), '')

        # basic decimal example
        value = decimal.Decimal('42.00')
        self.assertEqual(grid.render_currency(obj, 'foo', value), '$42.00')

        # basic float example
        value = 42.00
        self.assertEqual(grid.render_currency(obj, 'foo', value), '$42.00')

        # decimal places will be rounded
        value = decimal.Decimal('42.12345')
        self.assertEqual(grid.render_currency(obj, 'foo', value), '$42.12')

        # negative numbers get parens
        value = decimal.Decimal('-42.42')
        self.assertEqual(grid.render_currency(obj, 'foo', value), '($42.42)')

    def test_render_enum(self):
        enum = self.app.enum
        grid = self.make_grid(columns=['foo', 'bar'])
        obj = {'status': None}

        # null
        value = grid.render_enum(obj, 'status', None, enum=enum.UpgradeStatus)
        self.assertIsNone(value)

        # normal
        obj['status'] = enum.UpgradeStatus.SUCCESS
        value = grid.render_enum(obj, 'status', 'SUCCESS', enum=enum.UpgradeStatus)
        self.assertEqual(value, 'success')

    def test_render_percent(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        obj = MagicMock()

        # null
        self.assertEqual(grid.render_percent(obj, 'foo', None), "")

        # typical
        self.assertEqual(grid.render_percent(obj, 'foo', 12.3419), '12.34 %')

        # more decimal places
        self.assertEqual(grid.render_percent(obj, 'foo', 12.3419, decimals=3), '12.342 %')
        self.assertEqual(grid.render_percent(obj, 'foo', 12.3419, decimals=4), '12.3419 %')

        # negative
        self.assertEqual(grid.render_percent(obj, 'foo', -12.3419), '(12.34 %)')
        self.assertEqual(grid.render_percent(obj, 'foo', -12.3419, decimals=3), '(12.342 %)')

    def test_render_quantity(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        obj = MagicMock()

        # null
        self.assertEqual(grid.render_quantity(obj, 'foo', None), "")

        # integer decimals become integers
        value = decimal.Decimal('1.000')
        self.assertEqual(grid.render_quantity(obj, 'foo', value), "1")

        # but decimal places are preserved
        value = decimal.Decimal('1.234')
        self.assertEqual(grid.render_quantity(obj ,'foo', value), "1.234")

        # zero is *not* empty string (with this renderer)
        self.assertEqual(grid.render_quantity(obj, 'foo', 0), "0")

    def test_render_date(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        # null
        obj = MagicMock(dt=None)
        result = grid.render_date(obj, 'dt', None)
        self.assertEqual(result, '')

        # typical
        dt = datetime.date(2025, 1, 13)
        obj = MagicMock(dt=dt)
        result = grid.render_date(obj, 'dt', str(dt))
        self.assertEqual(result, '2025-01-13')

    def test_render_datetime(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        obj = MagicMock(dt=None)
        result = grid.render_datetime(obj, 'dt', None)
        self.assertEqual(result, '')

        dt = datetime.datetime(2024, 12, 12, 13, 44, tzinfo=datetime.timezone.utc)
        obj = MagicMock(dt=dt)
        result = grid.render_datetime(obj, 'dt', str(dt))
        self.assertEqual(result, '2024-12-12 13:44+0000')

    def test_render_vue_tag(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        html = grid.render_vue_tag()
        self.assertEqual(html, '<wutta-grid></wutta-grid>')

    def test_render_vue_template(self):
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')

        grid = self.make_grid(columns=['foo', 'bar'])
        html = grid.render_vue_template()
        self.assertIn('<script type="text/x-template" id="wutta-grid-template">', html)

    def test_render_table_element(self):
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')

        grid = self.make_grid(key='foobar', columns=['foo', 'bar'])

        # form not required
        html = grid.render_table_element()
        self.assertNotIn('<script ', html)
        self.assertIn('<b-table ', html)

        # form will register grid data
        form = Form(self.request)
        self.assertEqual(len(form.grid_vue_context), 0)
        html = grid.render_table_element(form)
        self.assertEqual(len(form.grid_vue_context), 1)
        self.assertIn('foobar', form.grid_vue_context)

    def test_render_vue_finalize(self):
        grid = self.make_grid()
        html = grid.render_vue_finalize()
        self.assertIn('<script>', html)
        self.assertIn("Vue.component('wutta-grid', WuttaGrid)", html)

    def test_get_vue_columns(self):

        # error if no columns are set
        grid = self.make_grid()
        self.assertRaises(ValueError, grid.get_vue_columns)

        # otherwise get back field/label dicts
        grid = self.make_grid(columns=['foo', 'bar'])
        columns = grid.get_vue_columns()
        first = columns[0]
        self.assertEqual(first['field'], 'foo')
        self.assertEqual(first['label'], 'Foo')

    def test_get_vue_active_sorters(self):
        model = self.app.model

        # empty
        grid = self.make_grid(key='foo', sortable=True, sort_on_backend=True)
        grid.load_settings()
        sorters = grid.get_vue_active_sorters()
        self.assertEqual(sorters, [])

        # format is different
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True,
                              sort_defaults='name')
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])
        sorters = grid.get_vue_active_sorters()
        self.assertEqual(sorters, [{'field': 'name', 'order': 'asc'}])

    def test_get_vue_first_sorter(self):

        # empty by default
        grid = self.make_grid(key='foo', sortable=True)
        sorter = grid.get_vue_first_sorter()
        self.assertIsNone(sorter)

        # will use first element from sort_defaults when applicable...

        # basic
        grid = self.make_grid(key='foo', sortable=True, sort_defaults='name')
        sorter = grid.get_vue_first_sorter()
        self.assertEqual(sorter, ['name', 'asc'])

        # descending
        grid = self.make_grid(key='foo', sortable=True, sort_defaults=('name', 'desc'))
        sorter = grid.get_vue_first_sorter()
        self.assertEqual(sorter, ['name', 'desc'])

        # multiple
        grid = self.make_grid(key='foo', sortable=True, sort_defaults=[('key', 'asc'), ('name', 'asc')])
        sorter = grid.get_vue_first_sorter()
        self.assertEqual(sorter, ['key', 'asc'])

        # will use first element from active_sorters when applicable...

        # basic
        grid = self.make_grid(key='foo', sortable=True)
        grid.active_sorters = [{'key': 'name', 'dir': 'asc'}]
        sorter = grid.get_vue_first_sorter()
        self.assertEqual(sorter, ['name', 'asc'])

        # descending
        grid = self.make_grid(key='foo', sortable=True)
        grid.active_sorters = [{'key': 'name', 'dir': 'desc'}]
        sorter = grid.get_vue_first_sorter()
        self.assertEqual(sorter, ['name', 'desc'])

        # multiple
        grid = self.make_grid(key='foo', sortable=True)
        grid.active_sorters = [{'key': 'key', 'dir': 'asc'}, {'key': 'name', 'dir': 'asc'}]
        sorter = grid.get_vue_first_sorter()
        self.assertEqual(sorter, ['key', 'asc'])

    def test_get_vue_filters(self):
        model = self.app.model

        # basic
        grid = self.make_grid(key='settings', model_class=model.Setting, filterable=True)
        grid.load_settings()
        filters = grid.get_vue_filters()
        self.assertEqual(len(filters), 2)
        name, value = filters
        self.assertEqual(name['choices'], [])
        self.assertEqual(name['choice_labels'], {})
        self.assertEqual(value['choices'], [])
        self.assertEqual(value['choice_labels'], {})

        class MockEnum(Enum):
            FOO = 'foo'
            BAR = 'bar'

        # with filter choices
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True, enums={'name': MockEnum})
        grid.load_settings()
        filters = grid.get_vue_filters()
        self.assertEqual(len(filters), 2)
        name, value = filters
        self.assertEqual(name['choices'], ['FOO', 'BAR'])
        self.assertEqual(name['choice_labels'], {'FOO': 'foo', 'BAR': 'bar'})
        self.assertEqual(value['choices'], [])
        self.assertEqual(value['choice_labels'], {})

    def test_object_to_dict(self):
        grid = self.make_grid()
        setting = {'name': 'foo', 'value': 'bar'}

        # new dict but with same values
        dct = grid.object_to_dict(setting)
        self.assertIsInstance(dct, dict)
        self.assertIsNot(dct, setting)
        self.assertEqual(dct, setting)

        # random object, not iterable
        class MockSetting:
            def __init__(self, **kw):
                self.__dict__.update(kw)
        mock = MockSetting(**setting)
        dct = grid.object_to_dict(mock)
        self.assertIsInstance(dct, dict)
        self.assertEqual(dct, setting)

    def test_get_vue_context(self):

        # empty if no columns defined
        grid = self.make_grid()
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [], 'row_classes': {}})

        # typical data is a list
        mydata = [
            {'foo': 'bar'},
        ]
        grid = self.make_grid(columns=['foo'], data=mydata)
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [{'foo': 'bar'}], 'row_classes': {}})

        # if grid has actions, that list may be supplemented
        grid.actions.append(mod.GridAction(self.request, 'view', url='/blarg'))
        context = grid.get_vue_context()
        self.assertIsNot(context['data'], mydata)
        self.assertEqual(context, {'data': [{'foo': 'bar', '_action_url_view': '/blarg'}],
                                   'row_classes': {}})

        # can override value rendering
        grid.set_renderer('foo', lambda record, key, value: "blah blah")
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [{'foo': 'blah blah', '_action_url_view': '/blarg'}],
                                   'row_classes': {}})

        # can set row class
        grid.row_class = 'whatever'
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [{'foo': 'blah blah', '_action_url_view': '/blarg'}],
                                   'row_classes': {'0': 'whatever'}})

    def test_get_vue_data(self):

        # empty if no columns defined
        grid = self.make_grid()
        data = grid.get_vue_data()
        self.assertEqual(data, [])

        # typical data is a list
        mydata = [
            {'foo': 'bar'},
        ]
        grid = self.make_grid(columns=['foo'], data=mydata)
        data = grid.get_vue_data()
        self.assertEqual(data, [{'foo': 'bar'}])

        # if grid has actions, that list may be supplemented
        grid.actions.append(mod.GridAction(self.request, 'view', url='/blarg'))
        data = grid.get_vue_data()
        self.assertIsNot(data, mydata)
        self.assertEqual(data, [{'foo': 'bar', '_action_url_view': '/blarg'}])

        # can override value rendering
        grid.set_renderer('foo', lambda record, key, value: "blah blah")
        data = grid.get_vue_data()
        self.assertEqual(data, [{'foo': 'blah blah', '_action_url_view': '/blarg'}])

    def test_get_row_class(self):
        model = self.app.model
        user = model.User(username='barney', active=True)
        self.session.add(user)
        self.session.commit()
        data = dict(user)

        # null by default
        grid = self.make_grid()
        self.assertIsNone(grid.get_row_class(user, data, 1))

        # can use static class
        grid.row_class = 'foo'
        self.assertEqual(grid.get_row_class(user, data, 1), 'foo')

        # can use callable
        def status(u, d, i):
            if not u.active:
                return 'inactive'
        grid.row_class = status
        self.assertIsNone(grid.get_row_class(user, data, 1))
        user.active = False
        self.assertEqual(grid.get_row_class(user, data, 1), 'inactive')

    def test_get_vue_pager_stats(self):
        data = [
            {'foo': 1, 'bar': 1},
            {'foo': 2, 'bar': 2},
            {'foo': 3, 'bar': 3},
            {'foo': 4, 'bar': 4},
            {'foo': 5, 'bar': 5},
            {'foo': 6, 'bar': 6},
            {'foo': 7, 'bar': 7},
            {'foo': 8, 'bar': 8},
            {'foo': 9, 'bar': 9},
        ]

        grid = self.make_grid(columns=['foo', 'bar'], pagesize=4, page=2)
        grid.pager = grid.paginate_data(data)
        stats = grid.get_vue_pager_stats()
        self.assertEqual(stats['item_count'], 9)
        self.assertEqual(stats['items_per_page'], 4)
        self.assertEqual(stats['page'], 2)
        self.assertEqual(stats['first_item'], 5)
        self.assertEqual(stats['last_item'], 8)


class TestGridAction(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config, use_oruga=False)

    def make_action(self, key, **kwargs):
        return mod.GridAction(self.request, key, **kwargs)

    def test_render_icon(self):

        # icon is derived from key by default
        action = self.make_action('blarg')
        html = action.render_icon()
        self.assertIn('<i class="fas fa-blarg">', html)

        # oruga has different output
        self.request.use_oruga = True
        html = action.render_icon()
        self.assertIn('<o-icon icon="blarg">', html)

    def test_render_label(self):

        # label is derived from key by default
        action = self.make_action('blarg')
        label = action.render_label()
        self.assertEqual(label, "Blarg")

        # otherwise use what caller provides
        action = self.make_action('foo', label="Bar")
        label = action.render_label()
        self.assertEqual(label, "Bar")

    def test_render_icon_and_label(self):
        action = self.make_action('blarg')
        with patch.multiple(action,
                            render_icon=lambda: 'ICON',
                            render_label=lambda: 'LABEL'):
            html = action.render_icon_and_label()
            self.assertEqual('ICON LABEL', html)

    def test_get_url(self):
        obj = {'foo': 'bar'}

        # null by default
        action = self.make_action('blarg')
        url = action.get_url(obj)
        self.assertIsNone(url)

        # or can be "static"
        action = self.make_action('blarg', url='/foo')
        url = action.get_url(obj)
        self.assertEqual(url, '/foo')

        # or can be "dynamic"
        action = self.make_action('blarg', url=lambda o, i: '/yeehaw')
        url = action.get_url(obj)
        self.assertEqual(url, '/yeehaw')



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Static Assets

Note that (for now?) It is assumed that *all* (i.e. even custom) apps
will include this module somewhere during startup.  For instance this
happens within :func:`wuttaweb.app.main()`::

   pyramid_config.include('wuttaweb.static')

This allows for certain common assets to be available for all apps.

However, an attempt is being made to incorporate Fanstatic for use
with the built-in static assets.  It is possible the above mechanism
could be abandoned in the future.

So on the Fanstatic front, we currently have defined:

.. data:: img

   A :class:`fanstatic:fanstatic.Library` representing the ``img``
   static folder.

.. data:: favicon

   A :class:`fanstatic:fanstatic.Resource` representing the
   ``img/favicon.ico`` image file.

.. data:: logo

   A :class:`fanstatic:fanstatic.Resource` representing the
   ``img/logo.png`` image file.

.. data:: testing

   A :class:`fanstatic:fanstatic.Resource` representing the
   ``img/testing.png`` image file.
"""

from fanstatic import Library, Resource


# fanstatic img library
img = Library('wuttaweb_img', 'img')
favicon = Resource(img, 'favicon.ico')
# nb. mock out the renderers here, to appease fanstatic
logo = Resource(img, 'logo.png', renderer=True)
testing = Resource(img, 'testing.png', renderer=True)


# TODO: should consider deprecating this?
def includeme(config):
    config.add_static_view('wuttaweb', 'wuttaweb:static')



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for people
"""

import sqlalchemy as sa

from wuttjamaican.db.model import Person
from wuttaweb.views import MasterView


class PersonView(MasterView):
    """
    Master view for people.

    Default route prefix is ``people``.

    Notable URLs provided by this class:

    * ``/people/``
    * ``/people/new``
    * ``/people/XXX``
    * ``/people/XXX/edit``
    * ``/people/XXX/delete``
    """
    model_class = Person
    model_title_plural = "People"
    route_prefix = 'people'
    sort_defaults = 'full_name'
    has_autocomplete = True

    grid_columns = [
        'full_name',
        'first_name',
        'middle_name',
        'last_name',
    ]

    filter_defaults = {
        'full_name': {'active': True},
    }

    form_fields = [
        'full_name',
        'first_name',
        'middle_name',
        'last_name',
        'users',
    ]

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)

        # full_name
        g.set_link('full_name')

        # first_name
        g.set_link('first_name')

        # last_name
        g.set_link('last_name')

    def configure_form(self, f):
        """ """
        super().configure_form(f)
        person = f.model_instance

        # full_name
        if self.creating or self.editing:
            f.remove('full_name')

        # users
        if self.viewing:
            f.set_grid('users', self.make_users_grid(person))

    def make_users_grid(self, person):
        """
        Make and return the grid for the Users field.

        This grid is shown for the Users field when viewing a Person.

        :returns: Fully configured :class:`~wuttaweb.grids.base.Grid`
           instance.
        """
        model = self.app.model
        route_prefix = self.get_route_prefix()

        grid = self.make_grid(key=f'{route_prefix}.view.users',
                              model_class=model.User,
                              data=person.users,
                              columns=[
                                  'username',
                                  'active',
                              ])

        if self.request.has_perm('users.view'):
            url = lambda user, i: self.request.route_url('users.view', uuid=user.uuid)
            grid.add_action('view', icon='eye', url=url)
            grid.set_link('username')

        if self.request.has_perm('users.edit'):
            url = lambda user, i: self.request.route_url('users.edit', uuid=user.uuid)
            grid.add_action('edit', url=url)

        return grid

    def objectify(self, form):
        """ """
        person = super().objectify(form)

        # full_name
        person.full_name = self.app.make_full_name(person.first_name,
                                                   person.last_name)

        return person

    def autocomplete_query(self, term):
        """ """
        model = self.app.model
        session = self.Session()
        query = session.query(model.Person)
        criteria = [model.Person.full_name.ilike(f'%{word}%')
                    for word in term.split()]
        query = query.filter(sa.and_(*criteria))\
                     .order_by(model.Person.full_name)
        return query

    def view_profile(self, session=None):
        """ """
        person = self.get_instance(session=session)
        context = {
            'person': person,
            'instance': person,
        }
        return self.render_to_response('view_profile', context)

    def make_user(self):
        """ """
        self.request.session.flash("TODO: this feature is not yet supported", 'error')
        return self.redirect(self.request.get_referrer())

    @classmethod
    def defaults(cls, config):
        """ """

        # nb. Person may come from custom model
        wutta_config = config.registry.settings['wutta_config']
        app = wutta_config.get_app()
        cls.model_class = app.model.Person

        cls._defaults(config)
        cls._people_defaults(config)

    @classmethod
    def _people_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        instance_url_prefix = cls.get_instance_url_prefix()
        permission_prefix = cls.get_permission_prefix()

        # view profile
        config.add_route(f'{route_prefix}.view_profile',
                         f'{instance_url_prefix}/profile',
                         request_method='GET')
        config.add_view(cls, attr='view_profile',
                        route_name=f'{route_prefix}.view_profile',
                        permission=f'{permission_prefix}.view_profile')

        # make user for person
        config.add_route(f'{route_prefix}.make_user',
                         f'{url_prefix}/make-user',
                         request_method='POST')
        config.add_view(cls, attr='make_user',
                        route_name=f'{route_prefix}.make_user',
                        permission='users.create')


def defaults(config, **kwargs):
    base = globals()

    PersonView = kwargs.get('PersonView', base['PersonView'])
    PersonView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Progress Indicators
"""

from wuttjamaican.progress import ProgressBase

from beaker.session import Session as BeakerSession


def get_basic_session(request, **kwargs):
    """
    Create/get a "basic" Beaker session object.
    """
    kwargs['use_cookies'] = False
    return BeakerSession(request, **kwargs)


def get_progress_session(request, key, **kwargs):
    """
    Create/get a Beaker session object, to be used for progress.
    """
    kwargs['id'] = f'{request.session.id}.progress.{key}'
    return get_basic_session(request, **kwargs)


class SessionProgress(ProgressBase):
    """
    Progress indicator which uses Beaker session storage to track
    current status.

    This is a subclass of
    :class:`wuttjamaican:wuttjamaican.progress.ProgressBase`.

    A view callable can create one of these, and then pass it into
    :meth:`~wuttjamaican.app.AppHandler.progress_loop()` or similar.

    As the loop updates progress along the way, this indicator will
    update the Beaker session to match.

    Separately then, the client side can send requests for the
    :func:`~wuttaweb.views.progress.progress()` view, to fetch current
    status out of the Beaker session.

    :param request: Current :term:`request` object.

    :param key: Unique key for this progress indicator.  Used to
       distinguish progress indicators in the Beaker session.

    Note that in addition to
    :meth:`~wuttjamaican:wuttjamaican.progress.ProgressBase.update()`
    and
    :meth:`~wuttjamaican:wuttjamaican.progress.ProgressBase.finish()`
    this progres class has some extra attributes and methods:

    .. attribute:: success_msg

       Optional message to display to the user (via session flash)
       when the operation completes successfully.

    .. attribute:: success_url

       URL to which user should be redirected, once the operation
       completes.

    .. attribute:: error_url

       URL to which user should be redirected, if the operation
       encounters an error.  If not specified, will fall back to
       :attr:`success_url`.
    """

    def __init__(self, request, key, success_msg=None, success_url=None, error_url=None):
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.key = key
        self.success_msg = success_msg
        self.success_url = success_url
        self.error_url = error_url or self.success_url
        self.session = get_progress_session(request, key)
        self.clear()

    def __call__(self, message, maximum):
        self.clear()
        self.session['message'] = message
        self.session['maximum'] = maximum
        self.session['maximum_display'] = f'{maximum:,d}'
        self.session['value'] = 0
        self.session.save()
        return self

    def clear(self):
        """ """
        self.session.clear()
        self.session['complete'] = False
        self.session['error'] = False
        self.session.save()

    def update(self, value):
        """ """
        self.session.load()
        self.session['value'] = value
        self.session.save()

    def handle_error(self, error, error_url=None):
        """
        This should be called by the view code, within a try/catch
        block upon error.

        The session storage will be updated to reflect details of the
        error.  Next time client requests the progress status it will
        learn of the error and redirect the user.

        :param error: :class:`python:Exception` instance.

        :param error_url: Optional redirect URL; if not specified
           :attr:`error_url` is used.
        """
        self.session.load()
        self.session['error'] = True
        self.session['error_msg'] = self.app.render_error(error)
        self.session['error_url'] = error_url or self.error_url
        self.session.save()

    def handle_success(self, success_msg=None, success_url=None):
        """
        This should be called by the view code, when the long-running
        operation completes.

        The session storage will be updated to reflect the completed
        status.  Next time client requests the progress status it will
        discover it has completed, and redirect the user.

        :param success_msg: Optional message to display to the user
           (via session flash) when the operation completes
           successfully.  If not specified :attr:`success_msg` (or
           nothing) is used

        :param success_url: Optional redirect URL; if not specified
           :attr:`success_url` is used.
        """
        self.session.load()
        self.session['complete'] = True
        self.session['success_msg'] = success_msg or self.success_msg
        self.session['success_url'] = success_url or self.success_url
        self.session.save()



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Web Utilities
"""

import decimal
import importlib
import json
import logging
import uuid as _uuid
import warnings

import sqlalchemy as sa
from sqlalchemy import orm

import colander
from pyramid.renderers import get_renderer
from webhelpers2.html import HTML, tags

from wuttjamaican.util import resource_path
from wuttaweb.db import Session


log = logging.getLogger(__name__)


class FieldList(list):
    """
    Convenience wrapper for a form's field list.  This is a subclass
    of :class:`python:list`.

    You normally would not need to instantiate this yourself, but it
    is used under the hood for
    :attr:`~wuttaweb.forms.base.Form.fields` as well as
    :attr:`~wuttaweb.grids.base.Grid.columns`.
    """

    def insert_before(self, field, newfield):
        """
        Insert a new field, before an existing field.

        :param field: String name for the existing field.

        :param newfield: String name for the new field, to be inserted
           just before the existing ``field``.
        """
        if field in self:
            i = self.index(field)
            self.insert(i, newfield)
        else:
            log.warning("field '%s' not found, will append new field: %s",
                        field, newfield)
            self.append(newfield)

    def insert_after(self, field, newfield):
        """
        Insert a new field, after an existing field.

        :param field: String name for the existing field.

        :param newfield: String name for the new field, to be inserted
           just after the existing ``field``.
        """
        if field in self:
            i = self.index(field)
            self.insert(i + 1, newfield)
        else:
            log.warning("field '%s' not found, will append new field: %s",
                        field, newfield)
            self.append(newfield)

    def set_sequence(self, fields):
        """
        Sort the list such that it matches the same sequence as the
        given fields list.

        This does not add or remove any elements, it just
        (potentially) rearranges the internal list elements.
        Therefore you do not need to explicitly declare *all* fields;
        just the ones you care about.

        The resulting field list will have the requested fields in
        order, at the *beginning* of the list.  Any unrequested fields
        will remain in the same order as they were previously, but
        will be placed *after* the requested fields.

        :param fields: List of fields in the desired order.
        """
        unimportant = len(self) + 1

        def getkey(field):
            if field in fields:
                return fields.index(field)
            return unimportant

        self.sort(key=getkey)


def get_form_data(request):
    """
    Returns the effective form data for the given request.

    Mostly this is a convenience, which simply returns one of the
    following, depending on various attributes of the request.

    * :attr:`pyramid:pyramid.request.Request.POST`
    * :attr:`pyramid:pyramid.request.Request.json_body`
    """
    # nb. we prefer JSON only if no POST is present
    # TODO: this seems to work for our use case at least, but perhaps
    # there is a better way?  see also
    # https://docs.pylonsproject.org/projects/pyramid/en/latest/api/request.html#pyramid.request.Request.is_xhr
    if not request.POST and (
            getattr(request, 'is_xhr', False)
            or getattr(request, 'content_type', None) == 'application/json'):
        return request.json_body
    return request.POST


def get_libver(
        request,
        key,
        configured_only=False,
        default_only=False,
        prefix='wuttaweb',
):
    """
    Return the appropriate version string for the web resource library
    identified by ``key``.

    WuttaWeb makes certain assumptions about which libraries would be
    used on the frontend, and which versions for each would be used by
    default.  But it should also be possible to customize which
    versions are used, hence this function.

    Each library has a built-in default version but your config can
    override them, e.g.:

    .. code-block:: ini

       [wuttaweb]
       libver.bb_vue = 3.4.29

    :param request: Current request.

    :param key: Unique key for the library, as string.  Possibilities
       are the same as for :func:`get_liburl()`.

    :param configured_only: Pass ``True`` here if you only want the
       configured version and ignore the default version.

    :param default_only: Pass ``True`` here if you only want the
       default version and ignore the configured version.

    :param prefix: If specified, will override the prefix used for
       config lookups.

       .. warning::

          This ``prefix`` param is for backward compatibility and may
          be removed in the future.

    :returns: The appropriate version string, e.g. ``'1.2.3'`` or
       ``'latest'`` etc.  Can also return ``None`` in some cases.
    """
    config = request.wutta_config

    # nb. we prefer a setting to be named like:  wuttaweb.libver.vue
    # but for back-compat this also can work:    tailbone.libver.vue
    # and for more back-compat this can work:    wuttaweb.vue_version
    # however that compat only works for some of the settings...

    if not default_only:

        # nb. new/preferred setting
        version = config.get(f'wuttaweb.libver.{key}')
        if version:
            return version

        # fallback to caller-specified prefix
        if prefix != 'wuttaweb':
            version = config.get(f'{prefix}.libver.{key}')
            if version:
                warnings.warn(f"config for {prefix}.libver.{key} is deprecated; "
                              f"please set wuttaweb.libver.{key} instead",
                              DeprecationWarning)
                return version

    if key == 'buefy':
        if not default_only:
            # nb. old/legacy setting
            version = config.get(f'{prefix}.buefy_version')
            if version:
                warnings.warn(f"config for {prefix}.buefy_version is deprecated; "
                              "please set wuttaweb.libver.buefy instead",
                              DeprecationWarning)
                return version
        if not configured_only:
            return '0.9.25'

    elif key == 'buefy.css':
        # nb. this always returns something
        return get_libver(request, 'buefy',
                          default_only=default_only,
                          configured_only=configured_only)

    elif key == 'vue':
        if not default_only:
            # nb. old/legacy setting
            version = config.get(f'{prefix}.vue_version')
            if version:
                warnings.warn(f"config for {prefix}.vue_version is deprecated; "
                              "please set wuttaweb.libver.vue instead",
                              DeprecationWarning)
                return version
        if not configured_only:
            return '2.6.14'

    elif key == 'vue_resource':
        if not configured_only:
            return '1.5.3'

    elif key == 'fontawesome':
        if not configured_only:
            return '5.3.1'

    elif key == 'bb_vue':
        if not configured_only:
            return '3.5.18'

    elif key == 'bb_oruga':
        if not configured_only:
            return '0.11.4'

    elif key in ('bb_oruga_bulma', 'bb_oruga_bulma_css'):
        if not configured_only:
            return '0.7.3'

    elif key == 'bb_fontawesome_svg_core':
        if not configured_only:
            return '7.0.0'

    elif key == 'bb_free_solid_svg_icons':
        if not configured_only:
            return '7.0.0'

    elif key == 'bb_vue_fontawesome':
        if not configured_only:
            return '3.1.1'


def get_liburl(
        request,
        key,
        configured_only=False,
        default_only=False,
        prefix='wuttaweb',
):
    """
    Return the appropriate URL for the web resource library identified
    by ``key``.

    WuttaWeb makes certain assumptions about which libraries would be
    used on the frontend, and which versions for each would be used by
    default.  But ultimately a URL must be determined for each, hence
    this function.

    Each library has a built-in default URL which references a public
    Internet (i.e. CDN) resource, but your config can override the
    final URL in two ways:

    The simplest way is to just override the *version* but otherwise
    let the default logic construct the URL.  See :func:`get_libver()`
    for more on that approach.

    The most flexible way is to override the URL explicitly, e.g.:

    .. code-block:: ini

       [wuttaweb]
       liburl.bb_vue = https://example.com/cache/vue-3.4.31.js

    :param request: Current request.

    :param key: Unique key for the library, as string.  Possibilities
       are:

       Vue 2 + Buefy

       * ``vue``
       * ``vue_resource``
       * ``buefy``
       * ``buefy.css``
       * ``fontawesome``

       Vue 3 + Oruga

       * ``bb_vue``
       * ``bb_oruga``
       * ``bb_oruga_bulma``
       * ``bb_oruga_bulma_css``
       * ``bb_fontawesome_svg_core``
       * ``bb_free_solid_svg_icons``
       * ``bb_vue_fontawesome``

    :param configured_only: Pass ``True`` here if you only want the
       configured URL and ignore the default URL.

    :param default_only: Pass ``True`` here if you only want the
       default URL and ignore the configured URL.

    :param prefix: If specified, will override the prefix used for
       config lookups.

       .. warning::

          This ``prefix`` param is for backward compatibility and may
          be removed in the future.

    :returns: The appropriate URL as string.  Can also return ``None``
       in some cases.
    """
    config = request.wutta_config

    if not default_only:

        # nb. new/preferred setting
        url = config.get(f'wuttaweb.liburl.{key}')
        if url:
            return url

        # fallback to caller-specified prefix
        url = config.get(f'{prefix}.liburl.{key}')
        if url:
            warnings.warn(f"config for {prefix}.liburl.{key} is deprecated; "
                          f"please set wuttaweb.liburl.{key} instead",
                          DeprecationWarning)
            return url

    if configured_only:
        return

    version = get_libver(request, key, prefix=prefix,
                         configured_only=False,
                         default_only=default_only)

    # load fanstatic libcache if configured
    static = config.get('wuttaweb.static_libcache.module')
    if not static:
        static = config.get(f'{prefix}.static_libcache.module')
        if static:
            warnings.warn(f"config for {prefix}.static_libcache.module is deprecated; "
                          "please set wuttaweb.static_libcache.module instead",
                          DeprecationWarning)
    if static:
        static = importlib.import_module(static)
        needed = request.environ['fanstatic.needed']
        liburl = needed.library_url(static.libcache) + '/'
        # nb. add custom url prefix if needed, e.g. /wutta
        if request.script_name:
            liburl = request.script_name + liburl

    if key == 'buefy':
        if static and hasattr(static, 'buefy_js'):
            return liburl + static.buefy_js.relpath
        return f'https://unpkg.com/buefy@{version}/dist/buefy.min.js'

    elif key == 'buefy.css':
        if static and hasattr(static, 'buefy_css'):
            return liburl + static.buefy_css.relpath
        return f'https://unpkg.com/buefy@{version}/dist/buefy.min.css'

    elif key == 'vue':
        if static and hasattr(static, 'vue_js'):
            return liburl + static.vue_js.relpath
        return f'https://unpkg.com/vue@{version}/dist/vue.min.js'

    elif key == 'vue_resource':
        if static and hasattr(static, 'vue_resource_js'):
            return liburl + static.vue_resource_js.relpath
        return f'https://cdn.jsdelivr.net/npm/vue-resource@{version}'

    elif key == 'fontawesome':
        if static and hasattr(static, 'fontawesome_js'):
            return liburl + static.fontawesome_js.relpath
        return f'https://use.fontawesome.com/releases/v{version}/js/all.js'

    elif key == 'bb_vue':
        if static and hasattr(static, 'bb_vue_js'):
            return liburl + static.bb_vue_js.relpath
        return f'https://unpkg.com/vue@{version}/dist/vue.esm-browser.prod.js'

    elif key == 'bb_oruga':
        if static and hasattr(static, 'bb_oruga_js'):
            return liburl + static.bb_oruga_js.relpath
        return f'https://unpkg.com/@oruga-ui/oruga-next@{version}/dist/oruga.mjs'

    elif key == 'bb_oruga_bulma':
        if static and hasattr(static, 'bb_oruga_bulma_js'):
            return liburl + static.bb_oruga_bulma_js.relpath
        return f'https://unpkg.com/@oruga-ui/theme-bulma@{version}/dist/bulma.js'

    elif key == 'bb_oruga_bulma_css':
        if static and hasattr(static, 'bb_oruga_bulma_css'):
            return liburl + static.bb_oruga_bulma_css.relpath
        return f'https://unpkg.com/@oruga-ui/theme-bulma@{version}/dist/bulma.css'

    elif key == 'bb_fontawesome_svg_core':
        if static and hasattr(static, 'bb_fontawesome_svg_core_js'):
            return liburl + static.bb_fontawesome_svg_core_js.relpath
        return f'https://cdn.jsdelivr.net/npm/@fortawesome/fontawesome-svg-core@{version}/+esm'

    elif key == 'bb_free_solid_svg_icons':
        if static and hasattr(static, 'bb_free_solid_svg_icons_js'):
            return liburl + static.bb_free_solid_svg_icons_js.relpath
        return f'https://cdn.jsdelivr.net/npm/@fortawesome/free-solid-svg-icons@{version}/+esm'

    elif key == 'bb_vue_fontawesome':
        if static and hasattr(static, 'bb_vue_fontawesome_js'):
            return liburl + static.bb_vue_fontawesome_js.relpath
        return f'https://cdn.jsdelivr.net/npm/@fortawesome/vue-fontawesome@{version}/+esm'


def get_csrf_token(request):
    """
    Convenience function, returns the effective CSRF token (raw
    string) for the given request.

    See also :func:`render_csrf_token()`.
    """
    token = request.session.get_csrf_token()
    if token is None:
        token = request.session.new_csrf_token()
    return token


def render_csrf_token(request, name='_csrf'):
    """
    Convenience function, returns CSRF hidden input inside hidden div,
    e.g.:

    .. code-block:: html

       <div style="display: none;">
          <input type="hidden" name="_csrf" value="TOKEN" />
       </div>

    This function is part of :mod:`wuttaweb.helpers` (as
    :func:`~wuttaweb.helpers.csrf_token()`) which means you can do
    this in page templates:

    .. code-block:: mako

       ${h.form(request.current_route_url())}
       ${h.csrf_token(request)}
       <!-- other fields etc. -->
       ${h.end_form()}

    See also :func:`get_csrf_token()`.
    """
    token = get_csrf_token(request)
    return HTML.tag('div', tags.hidden(name, value=token, id=None), style='display:none;')


def get_model_fields(config, model_class, include_fk=False):
    """
    Convenience function to return a list of field names for the given
    :term:`data model` class.

    This logic only supports SQLAlchemy mapped classes and will use
    that to determine the field listing if applicable.  Otherwise this
    returns ``None``.

    :param config: App :term:`config object`.

    :param model_class: Data model class.

    :param include_fk: Whether to include foreign key column names in
       the result.  They are excluded by default, since the
       relationship names are also included and generally preferred.

    :returns: List of field names, or ``None`` if it could not be
       determined.
    """
    try:
        mapper = sa.inspect(model_class)
    except sa.exc.NoInspectionAvailable:
        return

    if include_fk:
        fields = [prop.key for prop in mapper.iterate_properties]
    else:
        fields = [prop.key for prop in mapper.iterate_properties
                  if not prop_is_fk(mapper, prop)]

    # nb. we never want the continuum 'versions' prop
    app = config.get_app()
    if app.continuum_is_enabled() and 'versions' in fields:
        fields.remove('versions')

    return fields


def prop_is_fk(mapper, prop):
    """ """
    if not isinstance(prop, orm.ColumnProperty):
        return False

    prop_columns = [col.name for col in prop.columns]
    for rel in mapper.relationships:
        rel_columns = [col.name for col in rel.local_columns]
        if rel_columns == prop_columns:
            return True

    return False


def make_json_safe(value, key=None, warn=True):
    """
    Convert a Python value as needed, to ensure it is compatible with
    :func:`python:json.dumps()`.

    :param value: Python value.

    :param key: Optional key for the value, if known.  This is used
       when logging warnings, if applicable.

    :param warn: Whether warnings should be logged if the value is not
       already JSON-compatible.

    :returns: A (possibly new) Python value which is guaranteed to be
       JSON-serializable.
    """

    # convert null => None
    if value is colander.null:
        return None

    elif isinstance(value, dict):
        # recursively convert dict
        parent = dict(value)
        for key, value in parent.items():
            parent[key] = make_json_safe(value, key=key, warn=warn)
        value = parent

    elif isinstance(value, list):
        # recursively convert list
        parent = list(value)
        for i, value in enumerate(parent):
            parent[i] = make_json_safe(value, key=key, warn=warn)
        value = parent

    elif isinstance(value, _uuid.UUID):
        # convert UUID to str
        value = value.hex

    elif isinstance(value, decimal.Decimal):
        # convert decimal to float
        value = float(value)

    # ensure JSON-compatibility, warn if problems
    try:
        json.dumps(value)
    except TypeError as error:
        if warn:
            prefix = "value"
            if key:
                prefix += f" for '{key}'"
            log.warning("%s is not json-friendly: %s", prefix, repr(value))
        value = str(value)
        if warn:
            log.warning("forced value to: %s", value)

    return value


##############################
# theme functions
##############################

def get_available_themes(config):
    """
    Returns the official list of theme names which are available for
    use in the app.  Privileged users may choose among these when
    changing the global theme.

    If config specifies a list, that will be honored.  Otherwise the
    default list is:  ``['default', 'butterfly']``

    Note that the 'default' theme is Vue 2 + Buefy, while 'butterfly'
    is Vue 3 + Oruga.

    You can specify via config by setting e.g.:

    .. code-block:: ini

       [wuttaweb]
       themes.keys = default, butterfly, my-other-one

    :param config: App :term:`config object`.
    """
    # get available list from config, if it has one
    available = config.get_list('wuttaweb.themes.keys',
                                default=['default', 'butterfly'])

    # sort the list by name
    available.sort()

    # make default theme the first option
    if 'default' in available:
        available.remove('default')
    available.insert(0, 'default')

    return available


def get_effective_theme(config, theme=None, session=None):
    """
    Validate and return the "effective" theme.

    If caller specifies a ``theme`` then it will be returned (if
    "available" - see below).

    Otherwise the current theme will be read from db setting.  (Note
    we do not read simply from config object, we always read from db
    setting - this allows for the theme setting to change dynamically
    while app is running.)

    In either case if the theme is not listed in
    :func:`get_available_themes()` then a ``ValueError`` is raised.

    :param config: App :term:`config object`.

    :param theme: Optional name of desired theme, instead of getting
       current theme per db setting.

    :param session: Optional :term:`db session`.

    :returns: Name of theme.
    """
    app = config.get_app()

    if not theme:
        with app.short_session(session=session) as s:
            theme = app.get_setting(s, 'wuttaweb.theme') or 'default'

    # confirm requested theme is available
    available = get_available_themes(config)
    if theme not in available:
        raise ValueError(f"theme not available: {theme}")

    return theme


def get_theme_template_path(config, theme=None, session=None):
    """
    Return the template path for effective theme.

    If caller specifies a ``theme`` then it will be used; otherwise
    the current theme will be read from db setting.  The logic for
    that happens in :func:`get_effective_theme()`, which this function
    will call first.

    Once we have the valid theme name, we check config in case it
    specifies a template path override for it.  But if not, a default
    template path is assumed.

    The default path would be expected to live under
    ``wuttaweb:templates/themes``; for instance the ``butterfly``
    theme has a default template path of
    ``wuttaweb:templates/themes/butterfly``.

    :param config: App :term:`config object`.

    :param theme: Optional name of desired theme, instead of getting
       current theme per db setting.

    :param session: Optional :term:`db session`.

    :returns: Path on disk to theme template folder.
    """
    theme = get_effective_theme(config, theme=theme, session=session)
    theme_path = config.get(f'wuttaweb.theme.{theme}',
                            default=f'wuttaweb:templates/themes/{theme}')
    return resource_path(theme_path)


def set_app_theme(request, theme, session=None):
    """
    Set the effective theme for the running app.

    This will modify the *global* Mako template lookup directories,
    i.e. app templates will change for all users immediately.

    This will first validate the theme by calling
    :func:`get_effective_theme()`.  It then retrieves the template
    path via :func:`get_theme_template_path()`.

    The theme template path is then injected into the app settings
    registry such that it overrides the Mako lookup directories.

    It also will persist the theme name within db settings, so as to
    ensure it survives app restart.
    """
    config = request.wutta_config
    app = config.get_app()

    theme = get_effective_theme(config, theme=theme, session=session)
    theme_path = get_theme_template_path(config, theme=theme, session=session)

    # there's only one global template lookup; can get to it via any renderer
    # but should *not* use /base.mako since that one is about to get volatile
    renderer = get_renderer('/menu.mako')
    lookup = renderer.lookup

    # overwrite first entry in lookup's directory list
    lookup.directories[0] = theme_path

    # clear template cache for lookup object, so it will reload each (as needed)
    lookup._collection.clear()

    # persist current theme in db settings
    with app.short_session(session=session) as s:
        app.save_setting(s, 'wuttaweb.theme', theme)

    # and cache in live app settings
    request.registry.settings['wuttaweb.theme'] = theme



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaWeb - test utilities
"""

from unittest.mock import MagicMock

import fanstatic
from pyramid import testing

from wuttjamaican.testing import DataTestCase

from wuttaweb import subscribers


class WebTestCase(DataTestCase):
    """
    Base class for test suites requiring a full (typical) web app.
    """

    def setUp(self):
        self.setup_web()

    def setup_web(self):
        self.setup_db()
        self.request = self.make_request()
        self.pyramid_config = testing.setUp(request=self.request, settings={
            'wutta_config': self.config,
            'mako.directories': ['wuttaweb:templates'],
            'pyramid_deform.template_search_path': 'wuttaweb:templates/deform',
        })

        # init web
        self.pyramid_config.include('pyramid_deform')
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_directive('add_wutta_permission_group',
                                          'wuttaweb.auth.add_permission_group')
        self.pyramid_config.add_directive('add_wutta_permission',
                                          'wuttaweb.auth.add_permission')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')
        self.pyramid_config.include('wuttaweb.static')

        # nb. mock out fanstatic env..good enough for now to avoid errors..
        needed = fanstatic.init_needed()
        self.request.environ[fanstatic.NEEDED] = needed

        # setup new request w/ anonymous user
        event = MagicMock(request=self.request)
        subscribers.new_request(event)
        def user_getter(request, **kwargs): pass
        subscribers.new_request_set_user(event, db_session=self.session,
                                         user_getter=user_getter)

    def tearDown(self):
        self.teardown_web()

    def teardown_web(self):
        testing.tearDown()
        self.teardown_db()

    def make_request(self):
        return testing.DummyRequest()



# -*- coding: utf-8; -*-

from unittest.mock import patch, MagicMock

import pytest

from wuttaweb.db import continuum as mod
from wuttaweb.testing import WebTestCase


class TestWuttaWebContinuumPlugin(WebTestCase):

    def setUp(self):
        if not hasattr(mod, 'WuttaWebContinuumPlugin'):
            pytest.skip("test not relevant without sqlalchemy-continuum")
        self.setup_web()

    def make_plugin(self):
        return mod.WuttaWebContinuumPlugin()

    def test_get_remote_addr(self):
        plugin = self.make_plugin()

        with patch.object(mod, 'get_current_request', return_value=None):
            self.assertIsNone(plugin.get_remote_addr(None, self.session))

        self.request.client_addr = '127.0.0.1'
        self.assertEqual(plugin.get_remote_addr(None, self.session), '127.0.0.1')

    def test_get_user_id(self):
        plugin = self.make_plugin()

        with patch.object(mod, 'get_current_request', return_value=None):
            self.assertIsNone(plugin.get_user_id(None, self.session))

        self.request.user = MagicMock(uuid='some-random-uuid')
        self.assertEqual(plugin.get_user_id(None, self.session), 'some-random-uuid')



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Application
"""

import logging
import os

from wuttjamaican.app import AppProvider
from wuttjamaican.conf import make_config

from asgiref.wsgi import WsgiToAsgi
from pyramid.config import Configurator

import wuttaweb.db
from wuttaweb.auth import WuttaSecurityPolicy
from wuttaweb.util import get_effective_theme, get_theme_template_path


log = logging.getLogger(__name__)


class WebAppProvider(AppProvider):
    """
    The :term:`app provider` for WuttaWeb.  This adds some methods to
    the :term:`app handler`, which are specific to web apps.  It also
    registers some :term:`email templates <email template>` for the
    app, etc.
    """
    email_modules = ['wuttaweb.emails']
    email_templates = ['wuttaweb:email-templates']

    def get_web_handler(self, **kwargs):
        """
        Get the configured "web" handler for the app.

        Specify a custom handler in your config file like this:

        .. code-block:: ini

           [wutta]
           web.handler_spec = poser.web.handler:PoserWebHandler

        :returns: Instance of :class:`~wuttaweb.handler.WebHandler`.
        """
        if 'web_handler' not in self.__dict__:
            spec = self.config.get(f'{self.appname}.web.handler_spec',
                                   default='wuttaweb.handler:WebHandler')
            self.web_handler = self.app.load_object(spec)(self.config)
        return self.web_handler


def make_wutta_config(settings, config_maker=None, **kwargs):
    """
    Make a WuttaConfig object from the given settings.

    Note that ``settings`` dict will (typically) correspond to the
    ``[app:main]`` section of your config file.

    Regardless, the ``settings`` must contain a special key/value
    which is needed to identify the location of the config file.
    Assuming the typical scenario then, your config file should have
    an entry like this:

    .. code-block:: ini

       [app:main]
       wutta.config = %(__file__)s

    The ``%(__file__)s`` is auto-replaced with the config file path,
    so ultimately ``settings`` would contain something like (at
    minimum)::

       {'wutta.config': '/path/to/config/file'}

    If this config file path cannot be discovered, an error is raised.
    """
    wutta_config = settings.get('wutta_config')
    if not wutta_config:

        # validate config file path
        path = settings.get('wutta.config')
        if not path or not os.path.exists(path):
            raise ValueError("Please set 'wutta.config' in [app:main] "
                             "section of config to the path of your "
                             "config file.  Lame, but necessary.")

        # make config, add to settings
        config_maker = config_maker or make_config
        wutta_config = config_maker(path, **kwargs)
        settings['wutta_config'] = wutta_config

    # configure database sessions
    if hasattr(wutta_config, 'appdb_engine'):
        wuttaweb.db.Session.configure(bind=wutta_config.appdb_engine)

    return wutta_config


def make_pyramid_config(settings):
    """
    Make and return a Pyramid config object from the given settings.

    The config is initialized with certain features deemed useful for
    all apps.

    :returns: Instance of
       :class:`pyramid:pyramid.config.Configurator`.
    """
    settings.setdefault('fanstatic.versioning', 'true')
    settings.setdefault('mako.directories', ['wuttaweb:templates'])
    settings.setdefault('pyramid_deform.template_search_path',
                        'wuttaweb:templates/deform')

    # update settings per current theme
    establish_theme(settings)

    pyramid_config = Configurator(settings=settings)

    # configure user authorization / authentication
    pyramid_config.set_security_policy(WuttaSecurityPolicy())

    # require CSRF token for POST
    pyramid_config.set_default_csrf_options(require_csrf=True,
                                            token='_csrf',
                                            header='X-CSRF-TOKEN')

    pyramid_config.include('pyramid_beaker')
    pyramid_config.include('pyramid_deform')
    pyramid_config.include('pyramid_fanstatic')
    pyramid_config.include('pyramid_mako')
    pyramid_config.include('pyramid_tm')

    # add some permissions magic
    pyramid_config.add_directive('add_wutta_permission_group',
                                 'wuttaweb.auth.add_permission_group')
    pyramid_config.add_directive('add_wutta_permission',
                                 'wuttaweb.auth.add_permission')

    return pyramid_config


def main(global_config, **settings):
    """
    Make and return the WSGI application, per given settings.

    This function is designed to be called via Paste, hence it does
    require params and therefore can't be used directly as app factory
    for general WSGI servers.  For the latter see
    :func:`make_wsgi_app()` instead.

    And this *particular* function is not even that useful, it only
    constructs an app with minimal views built-in to WuttaWeb.  Most
    apps will define their own ``main()`` function (e.g.  as
    ``poser.web.app:main``), similar to this one but with additional
    views and other config.
    """
    wutta_config = make_wutta_config(settings)
    pyramid_config = make_pyramid_config(settings)

    pyramid_config.include('wuttaweb.static')
    pyramid_config.include('wuttaweb.subscribers')
    pyramid_config.include('wuttaweb.views')

    return pyramid_config.make_wsgi_app()


def make_wsgi_app(main_app=None, config=None):
    """
    Make and return a WSGI app, using the given Paste app factory.

    See also :func:`make_asgi_app()` for the ASGI equivalent.

    This function could be used directly for general WSGI servers
    (e.g. uvicorn), ***if*** you just want the built-in :func:`main()`
    app factory.

    But most likely you do not, in which case you must define your own
    function and call this one with your preferred app factory::

       from wuttaweb.app import make_wsgi_app

       def my_main(global_config, **settings):
           # TODO: build your app
           pass

       def make_my_wsgi_app():
           return make_wsgi_app(my_main)

    So ``make_my_wsgi_app()`` could then be used as-is for general
    WSGI servers.  However, note that this approach will require
    setting the ``WUTTA_CONFIG_FILES`` environment variable, unless
    running via :ref:`wutta-webapp`.

    :param main_app: Either a Paste-compatible app factory, or
       :term:`spec` for one.  If not specified, the built-in
       :func:`main()` is assumed.

    :param config: Optional :term:`config object`.  If not specified,
       one is created based on ``WUTTA_CONFIG_FILES`` environment
       variable.
    """
    if not config:
        config = make_config()
    app = config.get_app()

    # extract pyramid settings
    settings = config.get_dict('app:main')

    # keep same config object
    settings['wutta_config'] = config

    # determine the app factory
    if isinstance(main_app, str):
        make_wsgi_app = app.load_object(main_app)
    elif callable(main_app):
        make_wsgi_app = main_app
    else:
        raise ValueError("main_app must be spec or callable")

    # construct a pyramid app "per usual"
    return make_wsgi_app({}, **settings)


def make_asgi_app(main_app=None, config=None):
    """
    Make and return a ASGI app, using the given Paste app factory.

    This works the same as :func:`make_wsgi_app()` and should be
    called in the same way etc.
    """
    wsgi_app = make_wsgi_app(main_app, config=config)
    return WsgiToAsgi(wsgi_app)


def establish_theme(settings):
    """
    Establishes initial theme on app startup.  This mostly involves
    updating the given ``settings`` dict.

    This function is called automatically from within
    :func:`make_pyramid_config()`.

    It will first call :func:`~wuttaweb.util.get_effective_theme()` to
    read the current theme from the :term:`settings table`, and store
    this within ``settings['wuttaweb.theme']``.

    It then calls :func:`~wuttaweb.util.get_theme_template_path()` and
    will update ``settings['mako.directories']`` such that the theme's
    template path is listed first.
    """
    config = settings['wutta_config']

    theme = get_effective_theme(config)
    settings['wuttaweb.theme'] = theme

    directories = settings['mako.directories']
    if isinstance(directories, str):
        directories = config.parse_list(directories)

    path = get_theme_template_path(config)
    directories.insert(0, path)
    settings['mako.directories'] = directories
